/* Copyright 2017-2021 Xaptum, Inc.
 * Copyright 2021 NXP
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Functions in this file comes from Xaptum, https://github.com/xaptum/ecdaa
 * */

/* ************************************************************************** */
/* Includes                                                                   */
/* ************************************************************************** */
#include <stdio.h>
#include <string.h>
#include "nxEnsure.h"
#include <nxLog_App.h>

#include "ex_sss_fido_hash.h"
#include "ex_sss_fido_big_256_28.h"
#include "ex_sss_fido_ecp.h"

/* clang-format off */

/* ************************************************************************** */
/* Local Defines                                                              */
/* ************************************************************************** */

/* ************************************************************************** */
/* Structures and Typedefs                                                    */
/* ************************************************************************** */
#define MAXXES_FP256BN 24
#define FEXCESS_FP256BN ((int32_t)1<<MAXXES_FP256BN)	     /**< 2^(BASEBITS*NLEN-MODBITS) - normalised BIG can be multiplied by more than this before reduction */
#define ECP_FP256BN_LENGTH (2*MODBYTES_256_28 + 1)

/* ************************************************************************** */
/* Global Variables                                                           */
/* ************************************************************************** */
BIG_256_28 Modulus_FP256BN= {0xED33013,0x292DDBA,0x80A82D3,0x65FB129,0x49F0CDC,0x5EEE71A,0xD46E5F2,0xFFFCF0C,0xFFFFFFF,0xF};
BIG_256_28 CURVE_Cof_FP256BN= {0x1,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0};
BIG_256_28 R2modp_FP256BN= {0x3B9F8B,0xEDE3363,0xFEC54E8,0x92FFEE9,0x3C55F79,0x13C1C06,0xC0123FA,0xA12F2EA,0xE559B2A,0x8};
const chunk MConst_FP256BN= 0x537E5E5;
const int CURVE_A_FP256BN= 0;
const BIG_256_28 CURVE_B_FP256BN= {0x3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0};
const int CURVE_B_I_FP256BN= 3;


/* ************************************************************************** */
/* Static function declarations                                               */
/* ************************************************************************** */
static int logb2(uint32_t v);
static void FP_FP256BN_reduce(FP_FP256BN *a);
static int FP_FP256BN_iszilch(FP_FP256BN *x);
static void FP_FP256BN_one(FP_FP256BN *n);
static void FP_FP256BN_nres(FP_FP256BN *y,BIG_256_28 x);
static void FP_FP256BN_mod(BIG_256_28 a,DBIG_256_28 d);
static void FP_FP256BN_zero(FP_FP256BN *x);
static int FP_FP256BN_equals(FP_FP256BN *x,FP_FP256BN *y);
static void FP_FP256BN_redc(BIG_256_28 x,FP_FP256BN *y);
static void FP_FP256BN_copy(FP_FP256BN *y,FP_FP256BN *x);
static void FP_FP256BN_mul(FP_FP256BN *r,FP_FP256BN *a,FP_FP256BN *b);
static void FP_FP256BN_sqr(FP_FP256BN *r,FP_FP256BN *a);
static void FP_FP256BN_cmove(FP_FP256BN *a,FP_FP256BN *b,int d);
static void FP_FP256BN_neg(FP_FP256BN *r,FP_FP256BN *a);
static void FP_FP256BN_add(FP_FP256BN *r,FP_FP256BN *a,FP_FP256BN *b);
static void FP_FP256BN_norm(FP_FP256BN *x);
static void FP_FP256BN_rcopy(FP_FP256BN *y, const BIG_256_28 c);
static void FP_FP256BN_sub(FP_FP256BN *r,FP_FP256BN *a,FP_FP256BN *b);
static void FP_FP256BN_imul(FP_FP256BN *r,FP_FP256BN *a,int c);
static int teq(int32_t b,int32_t c);

/* ************************************************************************** */
/* Private Functions                                                          */
/* ************************************************************************** */

// https://graphics.stanford.edu/~seander/bithacks.html
// constant time log to base 2 (or number of bits in)

static int logb2(uint32_t v)
{
    int r;
    v |= v >> 1;
    v |= v >> 2;
    v |= v >> 4;
    v |= v >> 8;
    v |= v >> 16;

    v = v - ((v >> 1) & 0x55555555);
    v = (v & 0x33333333) + ((v >> 2) & 0x33333333);
    r = (((v + (v >> 4)) & 0xF0F0F0F) * 0x1010101) >> 24;
    return r;
}

/* Fully reduce a mod Modulus */
static void FP_FP256BN_reduce(FP_FP256BN *a)
{
    BIG_256_28 m;
    BIG_256_28_rcopy(m,Modulus_FP256BN);
    BIG_256_28_mod(a->g,m);
    a->XES=1;
}

/* test x==0 ? */
/* SU= 48 */
static int FP_FP256BN_iszilch(FP_FP256BN *x)
{
    BIG_256_28 m;
    BIG_256_28_rcopy(m,Modulus_FP256BN);
    BIG_256_28_mod(x->g,m);
    return BIG_256_28_iszilch(x->g);
}

/* set n=1 */
static void FP_FP256BN_one(FP_FP256BN *n)
{
    BIG_256_28 b;
    BIG_256_28_one(b);
    FP_FP256BN_nres(n,b);
}

/* convert to Montgomery n-residue form */
static void FP_FP256BN_nres(FP_FP256BN *y,BIG_256_28 x)
{
    DBIG_256_28 d;
    BIG_256_28 r;
    BIG_256_28_rcopy(r,R2modp_FP256BN);
    BIG_256_28_mul(d,x,r);
    FP_FP256BN_mod(y->g,d);
    y->XES=2;
}

/* reduce a DBIG to a BIG using Montgomery's no trial division method */
/* d is expected to be dnormed before entry */
/* SU= 112 */
static void FP_FP256BN_mod(BIG_256_28 a,DBIG_256_28 d)
{
    BIG_256_28 mdls;
    BIG_256_28_rcopy(mdls,Modulus_FP256BN);
    BIG_256_28_monty(a,mdls,MConst_FP256BN,d);
}

static void FP_FP256BN_zero(FP_FP256BN *x)
{
    BIG_256_28_zero(x->g);
    x->XES=1;
}

static int FP_FP256BN_equals(FP_FP256BN *x,FP_FP256BN *y)
{
    FP_FP256BN_reduce(x);
    FP_FP256BN_reduce(y);
    if (BIG_256_28_comp(x->g,y->g)==0) return 1;
    return 0;
}

/* convert back to regular form */
static void FP_FP256BN_redc(BIG_256_28 x,FP_FP256BN *y)
{
    DBIG_256_28 d;
    BIG_256_28_dzero(d);
    BIG_256_28_dscopy(d,y->g);
    FP_FP256BN_mod(x,d);
}

static void FP_FP256BN_copy(FP_FP256BN *y,FP_FP256BN *x)
{
    BIG_256_28_copy(y->g,x->g);
    y->XES=x->XES;
}

/* r=a*b mod Modulus */
/* product must be less that p.R - and we need to know this in advance! */
/* SU= 88 */
static void FP_FP256BN_mul(FP_FP256BN *r,FP_FP256BN *a,FP_FP256BN *b)
{
    DBIG_256_28 d;
//    chunk ea,eb;
//    BIG_256_28_norm(a);
//    BIG_256_28_norm(b);
//    ea=EXCESS_FP256BN(a->g);
//    eb=EXCESS_FP256BN(b->g);


    if ((int64_t)a->XES*b->XES>(int64_t)FEXCESS_FP256BN)
    {
        FP_FP256BN_reduce(a);  /* it is sufficient to fully reduce just one of them < p */
    }

    BIG_256_28_mul(d,a->g,b->g);
    FP_FP256BN_mod(r->g,d);

    r->XES=2;
}

/* Set r=a^2 mod m */
/* SU= 88 */
static void FP_FP256BN_sqr(FP_FP256BN *r,FP_FP256BN *a)
{
    DBIG_256_28 d;
//    chunk ea;
//    BIG_256_28_norm(a);
//    ea=EXCESS_FP256BN(a->g);


    if ((int64_t)a->XES*a->XES>(int64_t)FEXCESS_FP256BN)
    {
        FP_FP256BN_reduce(a);
    }

    BIG_256_28_sqr(d,a->g);
    FP_FP256BN_mod(r->g,d);
    r->XES=2;
}

/* Move b to a if d=1 */
static void FP_FP256BN_cmove(FP_FP256BN *a,FP_FP256BN *b,int d)
{
    int32_t c=-d;

    BIG_256_28_cmove(a->g,b->g,d);
    a->XES^=(a->XES^b->XES)&c;
}

/* Set r=-a mod Modulus */
/* SU= 64 */
static void FP_FP256BN_neg(FP_FP256BN *r,FP_FP256BN *a)
{
    int sb;
    BIG_256_28 m;

    BIG_256_28_rcopy(m,Modulus_FP256BN);

    sb=logb2(a->XES-1);
    BIG_256_28_fshl(m,sb);
    BIG_256_28_sub(r->g,m,a->g);
    r->XES=((int32_t)1<<sb);

    if (r->XES>FEXCESS_FP256BN)
    {
        FP_FP256BN_reduce(r);
    }

}

/* SU= 16 */
/* Set r=a+b */
static void FP_FP256BN_add(FP_FP256BN *r,FP_FP256BN *a,FP_FP256BN *b)
{
    BIG_256_28_add(r->g,a->g,b->g);
    r->XES=a->XES+b->XES;
    if (r->XES>FEXCESS_FP256BN)
    {
        FP_FP256BN_reduce(r);
    }
}

static void FP_FP256BN_norm(FP_FP256BN *x)
{
    BIG_256_28_norm(x->g);
}

static void FP_FP256BN_rcopy(FP_FP256BN *y, const BIG_256_28 c)
{
    BIG_256_28 b;
    BIG_256_28_rcopy(b,c);
    FP_FP256BN_nres(y,b);
}

/* Set r=a-b mod m */
/* SU= 56 */
static void FP_FP256BN_sub(FP_FP256BN *r,FP_FP256BN *a,FP_FP256BN *b)
{
    FP_FP256BN n;
//	BIG_256_28_norm(b);
    FP_FP256BN_neg(&n,b);
//	BIG_256_28_norm(n);
    FP_FP256BN_add(r,a,&n);
}

/* multiplication by an integer, r=a*c */
/* SU= 136 */
static void FP_FP256BN_imul(FP_FP256BN *r,FP_FP256BN *a,int c)
{
    int s=0;

    if (c<0)
    {
        c=-c;
        s=1;
    }

    //Montgomery
    BIG_256_28 k;
    FP_FP256BN f;
    if (a->XES*c<=FEXCESS_FP256BN)
    {
        BIG_256_28_pmul(r->g,a->g,c);
        r->XES=a->XES*c;    // careful here - XES jumps!
    }
    else
    {
        // don't want to do this - only a problem for Montgomery modulus and larger constants
        BIG_256_28_zero(k);
        BIG_256_28_inc(k,c);
        FP_FP256BN_nres(&f,k);
        FP_FP256BN_mul(r,a,&f);
    }
    /*
        if (c<=NEXCESS_256_28 && a->XES*c <= FEXCESS_FP256BN)
    	{
            BIG_256_28_imul(r->g,a->g,c);
    		r->XES=a->XES*c;
    		FP_FP256BN_norm(r);
    	}
        else
        {
                BIG_256_28_pxmul(d,a->g,c);

                BIG_256_28_rcopy(m,Modulus_FP256BN);
    			BIG_256_28_dmod(r->g,d,m);
                //FP_FP256BN_mod(r->g,d);                /// BIG problem here! Too slow for PM, How to do fast for Monty?
    			r->XES=2;
        }
    */
    if (s)
    {
        FP_FP256BN_neg(r,r);
        FP_FP256BN_norm(r);
    }
}

/* Conditional move Q to P dependant on d */
void ECP_FP256BN_cmove(ECP_FP256BN *P,ECP_FP256BN *Q,int d)
{
    FP_FP256BN_cmove(&(P->x),&(Q->x),d);
    FP_FP256BN_cmove(&(P->y),&(Q->y),d);
    FP_FP256BN_cmove(&(P->z),&(Q->z),d);

    d=~(d-1);
    P->inf^=(P->inf^Q->inf)&d;
}

/* Set P=O */
void ECP_FP256BN_inf(ECP_FP256BN *P)
{
    FP_FP256BN_zero(&(P->x));
    FP_FP256BN_one(&(P->y));
    FP_FP256BN_zero(&(P->z));
    P->inf=1;
}

/* test for P=O point-at-infinity */
int ECP_FP256BN_isinf(ECP_FP256BN *P)
{
    if (P->inf) return 1;
    FP_FP256BN_reduce(&(P->x));
    FP_FP256BN_reduce(&(P->z));
    FP_FP256BN_reduce(&(P->y));
    P->inf= (FP_FP256BN_iszilch(&(P->x)) && FP_FP256BN_iszilch(&(P->z)));
    return P->inf;
}

/* Convert P to Affine, from (x,y,z) to (x,y) */
/* SU=160 */
void ECP_FP256BN_affine(ECP_FP256BN *P)
{
    FP_FP256BN one,iz;
    BIG_256_28 m,b;
    if (ECP_FP256BN_isinf(P)) return;
    FP_FP256BN_one(&one);
    if (FP_FP256BN_equals(&(P->z),&one)) return;

    BIG_256_28_rcopy(m,Modulus_FP256BN);

    FP_FP256BN_redc(b,&(P->z));
    BIG_256_28_invmodp(b,b,m);
    FP_FP256BN_nres(&iz,b);

    FP_FP256BN_mul(&(P->x),&(P->x),&iz);

    FP_FP256BN_mul(&(P->y),&(P->y),&iz);
    FP_FP256BN_reduce(&(P->y));

    FP_FP256BN_reduce(&(P->x));
    FP_FP256BN_copy(&(P->z),&one);
}

/* Extract (x,y) and return sign of y. If x and y are the same return only x */
/* SU=16 */
int ECP_FP256BN_get(BIG_256_28 x,BIG_256_28 y,ECP_FP256BN *P)
{
    int s;

    if (ECP_FP256BN_isinf(P)) return -1;

    ECP_FP256BN_affine(P);

    FP_FP256BN_redc(y,&(P->y));
    s=BIG_256_28_parity(y);

    FP_FP256BN_redc(x,&(P->x));

    return s;
}

/* Set P=(x,{y}) */
/* SU=96 */
int ECP_FP256BN_set(ECP_FP256BN *P,BIG_256_28 x,BIG_256_28 y)
{
    FP_FP256BN rhs,y2;

    FP_FP256BN_nres(&y2,y);
    FP_FP256BN_sqr(&y2,&y2);
    FP_FP256BN_reduce(&y2);

    FP_FP256BN_nres(&rhs,x);
    ECP_FP256BN_rhs(&rhs,&rhs);

    if (!FP_FP256BN_equals(&y2,&rhs))
    {
        ECP_FP256BN_inf(P);
        return 0;
    }

    P->inf=0;

    FP_FP256BN_nres(&(P->x),x);
    FP_FP256BN_nres(&(P->y),y);
    FP_FP256BN_one(&(P->z));
    return 1;
}

/* Set P=Q */
void ECP_FP256BN_copy(ECP_FP256BN *P,ECP_FP256BN *Q)
{
    P->inf=Q->inf;
    FP_FP256BN_copy(&(P->x),&(Q->x));
    FP_FP256BN_copy(&(P->y),&(Q->y));
    FP_FP256BN_copy(&(P->z),&(Q->z));
}

/* Set P=r*P */
/* SU=424 */
void ECP_FP256BN_mul(ECP_FP256BN *P,BIG_256_28 e)
{
    /* fixed size windows */
    int i,nb,s,ns;
    BIG_256_28 mt,t;
    ECP_FP256BN Q,W[8],C;
    int8_t w[1+(NLEN_256_28*BASEBITS_256_28+3)/4];

    if (ECP_FP256BN_isinf(P)) return;
    if (BIG_256_28_iszilch(e))
    {
        ECP_FP256BN_inf(P);
        return;
    }

    ECP_FP256BN_affine(P);

    /* precompute table */

    ECP_FP256BN_copy(&Q,P);
    ECP_FP256BN_dbl(&Q);

    ECP_FP256BN_copy(&W[0],P);

    for (i=1; i<8; i++)
    {
        ECP_FP256BN_copy(&W[i],&W[i-1]);
        ECP_FP256BN_add(&W[i],&Q);
    }

//printf("W[1]= ");ECP_output(&W[1]); printf("\n");

    /* make exponent odd - add 2P if even, P if odd */
    BIG_256_28_copy(t,e);
    s=BIG_256_28_parity(t);
    BIG_256_28_inc(t,1);
    BIG_256_28_norm(t);
    ns=BIG_256_28_parity(t);
    BIG_256_28_copy(mt,t);
    BIG_256_28_inc(mt,1);
    BIG_256_28_norm(mt);
    BIG_256_28_cmove(t,mt,s);
    ECP_FP256BN_cmove(&Q,P,ns);
    ECP_FP256BN_copy(&C,&Q);

    nb=1+(BIG_256_28_nbits(t)+3)/4;

    /* convert exponent to signed 4-bit window */
    for (i=0; i<nb; i++)
    {
        w[i]=BIG_256_28_lastbits(t,5)-16;
        BIG_256_28_dec(t,w[i]);
        BIG_256_28_norm(t);
        BIG_256_28_fshr(t,4);
    }
    w[nb]=BIG_256_28_lastbits(t,5);

    ECP_FP256BN_copy(P,&W[(w[nb]-1)/2]);
    for (i=nb-1; i>=0; i--)
    {
        ECP_FP256BN_select(&Q,W,w[i]);
        ECP_FP256BN_dbl(P);
        ECP_FP256BN_dbl(P);
        ECP_FP256BN_dbl(P);
        ECP_FP256BN_dbl(P);
        ECP_FP256BN_add(P,&Q);
    }
    ECP_FP256BN_sub(P,&C); /* apply correction */
    ECP_FP256BN_affine(P);
}

/* Calculate right Hand Side of curve equation y^2=RHS */
/* SU=56 */
void ECP_FP256BN_rhs(FP_FP256BN *v,FP_FP256BN *x)
{
    /* x^3+Ax+B */
    FP_FP256BN t;
    FP_FP256BN_sqr(&t,x);
    FP_FP256BN_mul(&t,&t,x);

    if (CURVE_A_FP256BN==-3)
    {
        FP_FP256BN_neg(v,x);
        FP_FP256BN_norm(v);
        FP_FP256BN_imul(v,v,-CURVE_A_FP256BN);
        FP_FP256BN_norm(v);
        FP_FP256BN_add(v,&t,v);
    }
    else FP_FP256BN_copy(v,&t);

    FP_FP256BN_rcopy(&t,CURVE_B_FP256BN);

    FP_FP256BN_add(v,&t,v);
    FP_FP256BN_reduce(v);
}


/* Set P=2P */
/* SU=272 */
void ECP_FP256BN_dbl(ECP_FP256BN *P)
{
    FP_FP256BN t0,t1,t2,t3,x3,y3,z3,b;

    if (ECP_FP256BN_isinf(P)) return;

    if (CURVE_A_FP256BN==0)
    {
        //FP_FP256BN_copy(&t0,&(P->y));				//FP t0=new FP(y);
        FP_FP256BN_sqr(&t0,&(P->y));					//t0.sqr();
        //FP_FP256BN_copy(&t1,&(P->y));				//FP t1=new FP(y);
        FP_FP256BN_mul(&t1,&(P->y),&(P->z));			//t1.mul(z);

        //FP_FP256BN_copy(&t2,&(P->z));				//FP t2=new FP(z);
        FP_FP256BN_sqr(&t2,&(P->z));					//t2.sqr();

        //FP_FP256BN_copy(&(P->z),&t0);				//z.copy(t0);
        FP_FP256BN_add(&(P->z),&t0,&t0);		//z.add(t0);
        FP_FP256BN_norm(&(P->z));					//z.norm();
        FP_FP256BN_add(&(P->z),&(P->z),&(P->z));	//z.add(z);
        FP_FP256BN_add(&(P->z),&(P->z),&(P->z));	//z.add(z);
        FP_FP256BN_norm(&(P->z));					//z.norm();

        FP_FP256BN_imul(&t2,&t2,3*CURVE_B_I_FP256BN);		//t2.imul(3*ROM.CURVE_B_I);
        //FP_FP256BN_copy(&x3,&t2);					//FP x3=new FP(t2);
        FP_FP256BN_mul(&x3,&t2,&(P->z));			//x3.mul(z);

        //FP_FP256BN_copy(&y3,&t0);					//FP y3=new FP(t0);
        FP_FP256BN_add(&y3,&t0,&t2);				//y3.add(t2);
        FP_FP256BN_norm(&y3);						//y3.norm();
        FP_FP256BN_mul(&(P->z),&(P->z),&t1);		//z.mul(t1);

        //FP_FP256BN_copy(&t1,&t2);					//t1.copy(t2);
        FP_FP256BN_add(&t1,&t2,&t2);				//t1.add(t2);
        FP_FP256BN_add(&t2,&t2,&t1);				//t2.add(t1);
        FP_FP256BN_sub(&t0,&t0,&t2);				//t0.sub(t2);
        FP_FP256BN_norm(&t0);						//t0.norm();
        FP_FP256BN_mul(&y3,&y3,&t0);				//y3.mul(t0);
        FP_FP256BN_add(&y3,&y3,&x3);				//y3.add(x3);
        //FP_FP256BN_copy(&t1,&(P->x));				//t1.copy(x);                *** optimization possible
        FP_FP256BN_mul(&t1,&(P->x),&(P->y));			//t1.mul(y);
        //FP_FP256BN_copy(&(P->x),&t0);				//x.copy(t0);
        FP_FP256BN_norm(&t0);					//x.norm();
        FP_FP256BN_mul(&(P->x),&t0,&t1);		//x.mul(t1);
        FP_FP256BN_add(&(P->x),&(P->x),&(P->x));	//x.add(x);
        FP_FP256BN_norm(&(P->x));					//x.norm();
        FP_FP256BN_copy(&(P->y),&y3);				//y.copy(y3);
        FP_FP256BN_norm(&(P->y));					//y.norm();
    }
    else // its -3
    {
        //FP_FP256BN_copy(&t0,&(P->x));				//FP t0=new FP(x);
        //FP_FP256BN_copy(&t1,&(P->y));				//FP t1=new FP(y);
        //FP_FP256BN_copy(&t2,&(P->z));				//FP t2=new FP(z);
        //FP_FP256BN_copy(&t3,&(P->x));				//FP t3=new FP(x);
        //FP_FP256BN_copy(&z3,&(P->z));				//FP z3=new FP(z);

        if (CURVE_B_I_FP256BN==0)					//if (ROM.CURVE_B_I==0)
            FP_FP256BN_rcopy(&b,CURVE_B_FP256BN);		//b.copy(new FP(new BIG(ROM.CURVE_B)));

        FP_FP256BN_sqr(&t0,&(P->x));					//t0.sqr();  //1    x^2
        FP_FP256BN_sqr(&t1,&(P->y));					//t1.sqr();  //2    y^2
        FP_FP256BN_sqr(&t2,&(P->z));					//t2.sqr();  //3

        FP_FP256BN_mul(&t3,&(P->x),&(P->y));			//t3.mul(y); //4
        FP_FP256BN_add(&t3,&t3,&t3);				//t3.add(t3);
        FP_FP256BN_norm(&t3);						//t3.norm();//5

        FP_FP256BN_mul(&z3,&(P->z),&(P->x));			//z3.mul(x);   //6
        FP_FP256BN_add(&z3,&z3,&z3);				//z3.add(z3);
        FP_FP256BN_norm(&z3);						//z3.norm();//7
        //FP_FP256BN_copy(&y3,&t2);					//y3.copy(t2);

        if (CURVE_B_I_FP256BN==0)						//if (ROM.CURVE_B_I==0)
            FP_FP256BN_mul(&y3,&t2,&b);				//y3.mul(b); //8
        else
            FP_FP256BN_imul(&y3,&t2,CURVE_B_I_FP256BN);	//y3.imul(ROM.CURVE_B_I);

        FP_FP256BN_sub(&y3,&y3,&z3);				//y3.sub(z3); //y3.norm(); //9  ***
        //FP_FP256BN_copy(&x3,&y3);					//x3.copy(y3);
        FP_FP256BN_add(&x3,&y3,&y3);				//x3.add(y3);
        FP_FP256BN_norm(&x3);						//x3.norm();//10

        FP_FP256BN_add(&y3,&y3,&x3);				//y3.add(x3); //y3.norm();//11
        //FP_FP256BN_copy(&x3,&t1);					//x3.copy(t1);
        FP_FP256BN_sub(&x3,&t1,&y3);				//x3.sub(y3);
        FP_FP256BN_norm(&x3);						//x3.norm();//12
        FP_FP256BN_add(&y3,&y3,&t1);				//y3.add(t1);
        FP_FP256BN_norm(&y3);						//y3.norm();//13
        FP_FP256BN_mul(&y3,&y3,&x3);				//y3.mul(x3); //14
        FP_FP256BN_mul(&x3,&x3,&t3);				//x3.mul(t3); //15
        //FP_FP256BN_copy(&t3,&t2);					//t3.copy(t2);
        FP_FP256BN_add(&t3,&t2,&t2);				//t3.add(t2);  //16
        FP_FP256BN_add(&t2,&t2,&t3);				//t2.add(t3); //17

        if (CURVE_B_I_FP256BN==0)					//if (ROM.CURVE_B_I==0)
            FP_FP256BN_mul(&z3,&z3,&b);				//z3.mul(b); //18
        else
            FP_FP256BN_imul(&z3,&z3,CURVE_B_I_FP256BN);	//z3.imul(ROM.CURVE_B_I);

        FP_FP256BN_sub(&z3,&z3,&t2);				//z3.sub(t2); //z3.norm();//19
        FP_FP256BN_sub(&z3,&z3,&t0);				//z3.sub(t0);
        FP_FP256BN_norm(&z3);						//z3.norm();//20  ***
        //FP_FP256BN_copy(&t3,&z3);					//t3.copy(z3);
        FP_FP256BN_add(&t3,&z3,&z3);				//t3.add(z3); //t3.norm();//21

        FP_FP256BN_add(&z3,&z3,&t3);				//z3.add(t3);
        FP_FP256BN_norm(&z3);						//z3.norm(); //22
        //FP_FP256BN_copy(&t3,&t0);					//t3.copy(t0);
        FP_FP256BN_add(&t3,&t0,&t0);				//t3.add(t0); //t3.norm(); //23
        FP_FP256BN_add(&t0,&t0,&t3);				//t0.add(t3); //t0.norm();//24
        FP_FP256BN_sub(&t0,&t0,&t2);				//t0.sub(t2);
        FP_FP256BN_norm(&t0);						//t0.norm();//25

        FP_FP256BN_mul(&t0,&t0,&z3);				//t0.mul(z3);//26
        FP_FP256BN_add(&y3,&y3,&t0);				//y3.add(t0); //y3.norm();//27
        //FP_FP256BN_copy(&t0,&(P->y));				//t0.copy(y);
        FP_FP256BN_mul(&t0,&(P->y),&(P->z));			//t0.mul(z);//28
        FP_FP256BN_add(&t0,&t0,&t0);				//t0.add(t0);
        FP_FP256BN_norm(&t0);						//t0.norm(); //29
        FP_FP256BN_mul(&z3,&z3,&t0);				//z3.mul(t0);//30
        FP_FP256BN_sub(&(P->x),&x3,&z3);				//x3.sub(z3); //x3.norm();//31
        FP_FP256BN_add(&t0,&t0,&t0);				//t0.add(t0);
        FP_FP256BN_norm(&t0);						//t0.norm();//32
        FP_FP256BN_add(&t1,&t1,&t1);				//t1.add(t1);
        FP_FP256BN_norm(&t1);						//t1.norm();//33
        //FP_FP256BN_copy(&z3,&t0);					//z3.copy(t0);
        FP_FP256BN_mul(&(P->z),&t0,&t1);				//z3.mul(t1);//34

        //FP_FP256BN_copy(&(P->x),&x3);				//x.copy(x3);
        FP_FP256BN_norm(&(P->x));					//x.norm();
        FP_FP256BN_copy(&(P->y),&y3);				//y.copy(y3);
        FP_FP256BN_norm(&(P->y));					//y.norm();
        //FP_FP256BN_copy(&(P->z),&z3);				//z.copy(z3);
        FP_FP256BN_norm(&(P->z));					//z.norm();
    }
}

/* Set P+=Q */
/* SU=248 */
void ECP_FP256BN_add(ECP_FP256BN *P,ECP_FP256BN *Q)
{
    int b3;
    FP_FP256BN t0,t1,t2,t3,t4,x3,y3,z3,b;

    if (ECP_FP256BN_isinf(Q)) return;
    if (ECP_FP256BN_isinf(P))
    {
        ECP_FP256BN_copy(P,Q);
        return;
    }
    if (CURVE_A_FP256BN==0)
    {
        b3=3*CURVE_B_I_FP256BN;					//int b=3*ROM.CURVE_B_I;
        //FP_FP256BN_copy(&t0,&(P->x));			//FP t0=new FP(x);
        FP_FP256BN_mul(&t0,&(P->x),&(Q->x));		//t0.mul(Q.x);
        //FP_FP256BN_copy(&t1,&(P->y));			//FP t1=new FP(y);
        FP_FP256BN_mul(&t1,&(P->y),&(Q->y));		//t1.mul(Q.y);
        //FP_FP256BN_copy(&t2,&(P->z));			//FP t2=new FP(z);
        FP_FP256BN_mul(&t2,&(P->z),&(Q->z));		//t2.mul(Q.z);
        //FP_FP256BN_copy(&t3,&(P->x));			//FP t3=new FP(x);
        FP_FP256BN_add(&t3,&(P->x),&(P->y));		//t3.add(y);
        FP_FP256BN_norm(&t3);					//t3.norm();
        //FP_FP256BN_copy(&t4,&(Q->x));			//FP t4=new FP(Q.x);
        FP_FP256BN_add(&t4,&(Q->x),&(Q->y));		//t4.add(Q.y);
        FP_FP256BN_norm(&t4);					//t4.norm();
        FP_FP256BN_mul(&t3,&t3,&t4);			//t3.mul(t4);
        //FP_FP256BN_copy(&t4,&t0);				//t4.copy(t0);
        FP_FP256BN_add(&t4,&t0,&t1);			//t4.add(t1);

        FP_FP256BN_sub(&t3,&t3,&t4);			//t3.sub(t4);
        FP_FP256BN_norm(&t3);					//t3.norm();
        //FP_FP256BN_copy(&t4,&(P->y));			//t4.copy(y);
        FP_FP256BN_add(&t4,&(P->y),&(P->z));		//t4.add(z);
        FP_FP256BN_norm(&t4);					//t4.norm();
        //FP_FP256BN_copy(&x3,&(Q->y));			//FP x3=new FP(Q.y);
        FP_FP256BN_add(&x3,&(Q->y),&(Q->z));		//x3.add(Q.z);
        FP_FP256BN_norm(&x3);					//x3.norm();

        FP_FP256BN_mul(&t4,&t4,&x3);			//t4.mul(x3);
        //FP_FP256BN_copy(&x3,&t1);				//x3.copy(t1);
        FP_FP256BN_add(&x3,&t1,&t2);			//x3.add(t2);

        FP_FP256BN_sub(&t4,&t4,&x3);			//t4.sub(x3);
        FP_FP256BN_norm(&t4);					//t4.norm();
        //FP_FP256BN_copy(&x3,&(P->x));			//x3.copy(x);
        FP_FP256BN_add(&x3,&(P->x),&(P->z));		//x3.add(z);
        FP_FP256BN_norm(&x3);					//x3.norm();
        //FP_FP256BN_copy(&y3,&(Q->x));			//FP y3=new FP(Q.x);
        FP_FP256BN_add(&y3,&(Q->x),&(Q->z));		//y3.add(Q.z);
        FP_FP256BN_norm(&y3);					//y3.norm();
        FP_FP256BN_mul(&x3,&x3,&y3);			//x3.mul(y3);
        //FP_FP256BN_copy(&y3,&t0);				//y3.copy(t0);
        FP_FP256BN_add(&y3,&t0,&t2);			//y3.add(t2);
        FP_FP256BN_sub(&y3,&x3,&y3);			//y3.rsub(x3);
        FP_FP256BN_norm(&y3);					//y3.norm();
        //FP_FP256BN_copy(&x3,&t0);				//x3.copy(t0);
        FP_FP256BN_add(&x3,&t0,&t0);			//x3.add(t0);
        FP_FP256BN_add(&t0,&t0,&x3);			//t0.add(x3);
        FP_FP256BN_norm(&t0);					//t0.norm();
        FP_FP256BN_imul(&t2,&t2,b3);				//t2.imul(b);

        //FP_FP256BN_copy(&z3,&t1);				//FP z3=new FP(t1);
        FP_FP256BN_add(&z3,&t1,&t2);			//z3.add(t2);
        FP_FP256BN_norm(&z3);					//z3.norm();
        FP_FP256BN_sub(&t1,&t1,&t2);			//t1.sub(t2);
        FP_FP256BN_norm(&t1);					//t1.norm();
        FP_FP256BN_imul(&y3,&y3,b3);				//y3.imul(b);

        //FP_FP256BN_copy(&x3,&y3);				//x3.copy(y3);
        FP_FP256BN_mul(&x3,&y3,&t4);			//x3.mul(t4);
        //FP_FP256BN_copy(&t2,&t3);				//t2.copy(t3);
        FP_FP256BN_mul(&t2,&t3,&t1);			//t2.mul(t1);
        FP_FP256BN_sub(&(P->x),&t2,&x3);			//x3.rsub(t2);
        FP_FP256BN_mul(&y3,&y3,&t0);			//y3.mul(t0);
        FP_FP256BN_mul(&t1,&t1,&z3);			//t1.mul(z3);
        FP_FP256BN_add(&(P->y),&y3,&t1);			//y3.add(t1);
        FP_FP256BN_mul(&t0,&t0,&t3);			//t0.mul(t3);
        FP_FP256BN_mul(&z3,&z3,&t4);			//z3.mul(t4);
        FP_FP256BN_add(&(P->z),&z3,&t0);			//z3.add(t0);

        //FP_FP256BN_copy(&(P->x),&x3);			//x.copy(x3);
        FP_FP256BN_norm(&(P->x));				//x.norm();
        //FP_FP256BN_copy(&(P->y),&y3);			//y.copy(y3);
        FP_FP256BN_norm(&(P->y));				//y.norm();
        //FP_FP256BN_copy(&(P->z),&z3);			//z.copy(z3);
        FP_FP256BN_norm(&(P->z));				//z.norm();
    }
    else
    {
        //FP_FP256BN_copy(&t0,&(P->x));			//FP t0=new FP(x);
        //FP_FP256BN_copy(&t1,&(P->y));			//FP t1=new FP(y);
        //FP_FP256BN_copy(&t2,&(P->z));			//FP t2=new FP(z);
        //FP_FP256BN_copy(&t3,&(P->x));			//FP t3=new FP(x);
        //FP_FP256BN_copy(&t4,&(Q->x));			//FP t4=new FP(Q.x);

        //FP_FP256BN_copy(&y3,&(Q->x));			//FP y3=new FP(Q.x);
        //FP_FP256BN_copy(&x3,&(Q->y));			//FP x3=new FP(Q.y);

        if (CURVE_B_I_FP256BN==0)				//if (ROM.CURVE_B_I==0)
            FP_FP256BN_rcopy(&b,CURVE_B_FP256BN);	//b.copy(new FP(new BIG(ROM.CURVE_B)));

        FP_FP256BN_mul(&t0,&(P->x),&(Q->x));		//t0.mul(Q.x); //1
        FP_FP256BN_mul(&t1,&(P->y),&(Q->y));		//t1.mul(Q.y); //2
        FP_FP256BN_mul(&t2,&(P->z),&(Q->z));		//t2.mul(Q.z); //3

        FP_FP256BN_add(&t3,&(P->x),&(P->y));		//t3.add(y);
        FP_FP256BN_norm(&t3);					//t3.norm(); //4
        FP_FP256BN_add(&t4,&(Q->x),&(Q->y));		//t4.add(Q.y);
        FP_FP256BN_norm(&t4);					//t4.norm();//5
        FP_FP256BN_mul(&t3,&t3,&t4);			//t3.mul(t4);//6
        //FP_FP256BN_copy(&t4,&t0);				//t4.copy(t0);
        FP_FP256BN_add(&t4,&t0,&t1);			//t4.add(t1); //t4.norm(); //7
        FP_FP256BN_sub(&t3,&t3,&t4);			//t3.sub(t4);
        FP_FP256BN_norm(&t3);					//t3.norm(); //8
        //FP_FP256BN_copy(&t4,&(P->y));			//t4.copy(y);
        FP_FP256BN_add(&t4,&(P->y),&(P->z));		//t4.add(z);
        FP_FP256BN_norm(&t4);					//t4.norm();//9
        FP_FP256BN_add(&x3,&(Q->y),&(Q->z));		//x3.add(Q.z);
        FP_FP256BN_norm(&x3);					//x3.norm();//10
        FP_FP256BN_mul(&t4,&t4,&x3);			//t4.mul(x3); //11
        //FP_FP256BN_copy(&x3,&t1);				//x3.copy(t1);
        FP_FP256BN_add(&x3,&t1,&t2);			//x3.add(t2); //x3.norm();//12

        FP_FP256BN_sub(&t4,&t4,&x3);			//t4.sub(x3);
        FP_FP256BN_norm(&t4);					//t4.norm();//13
        //FP_FP256BN_copy(&x3,&(P->x));			//x3.copy(x);
        FP_FP256BN_add(&x3,&(P->x),&(P->z));		//x3.add(z);
        FP_FP256BN_norm(&x3);					//x3.norm(); //14
        FP_FP256BN_add(&y3,&(Q->x),&(Q->z));		//y3.add(Q.z);
        FP_FP256BN_norm(&y3);					//y3.norm();//15

        FP_FP256BN_mul(&x3,&x3,&y3);			//x3.mul(y3); //16
        //FP_FP256BN_copy(&y3,&t0);				//y3.copy(t0);
        FP_FP256BN_add(&y3,&t0,&t2);			//y3.add(t2); //y3.norm();//17

        FP_FP256BN_sub(&y3,&x3,&y3);			//y3.rsub(x3);
        FP_FP256BN_norm(&y3);					//y3.norm(); //18
        //FP_FP256BN_copy(&z3,&t2);				//z3.copy(t2);

        if (CURVE_B_I_FP256BN==0)				//if (ROM.CURVE_B_I==0)
            FP_FP256BN_mul(&z3,&t2,&b);			//z3.mul(b); //18
        else
            FP_FP256BN_imul(&z3,&t2,CURVE_B_I_FP256BN);	//z3.imul(ROM.CURVE_B_I);

        //FP_FP256BN_copy(&x3,&y3);				//x3.copy(y3);
        FP_FP256BN_sub(&x3,&y3,&z3);			//x3.sub(z3);
        FP_FP256BN_norm(&x3);					//x3.norm(); //20
        //FP_FP256BN_copy(&z3,&x3);				//z3.copy(x3);
        FP_FP256BN_add(&z3,&x3,&x3);			//z3.add(x3); //z3.norm(); //21

        FP_FP256BN_add(&x3,&x3,&z3);			//x3.add(z3); //x3.norm(); //22
        //FP_FP256BN_copy(&z3,&t1);				//z3.copy(t1);
        FP_FP256BN_sub(&z3,&t1,&x3);			//z3.sub(x3);
        FP_FP256BN_norm(&z3);					//z3.norm(); //23
        FP_FP256BN_add(&x3,&x3,&t1);			//x3.add(t1);
        FP_FP256BN_norm(&x3);					//x3.norm(); //24

        if (CURVE_B_I_FP256BN==0)				//if (ROM.CURVE_B_I==0)
            FP_FP256BN_mul(&y3,&y3,&b);			//y3.mul(b); //18
        else
            FP_FP256BN_imul(&y3,&y3,CURVE_B_I_FP256BN);	//y3.imul(ROM.CURVE_B_I);

        //FP_FP256BN_copy(&t1,&t2);				//t1.copy(t2);
        FP_FP256BN_add(&t1,&t2,&t2);			//t1.add(t2); //t1.norm();//26
        FP_FP256BN_add(&t2,&t2,&t1);			//t2.add(t1); //t2.norm();//27

        FP_FP256BN_sub(&y3,&y3,&t2);			//y3.sub(t2); //y3.norm(); //28

        FP_FP256BN_sub(&y3,&y3,&t0);			//y3.sub(t0);
        FP_FP256BN_norm(&y3);					//y3.norm(); //29
        //FP_FP256BN_copy(&t1,&y3);				//t1.copy(y3);
        FP_FP256BN_add(&t1,&y3,&y3);			//t1.add(y3); //t1.norm();//30
        FP_FP256BN_add(&y3,&y3,&t1);			//y3.add(t1);
        FP_FP256BN_norm(&y3);					//y3.norm(); //31

        //FP_FP256BN_copy(&t1,&t0);				//t1.copy(t0);
        FP_FP256BN_add(&t1,&t0,&t0);			//t1.add(t0); //t1.norm(); //32
        FP_FP256BN_add(&t0,&t0,&t1);			//t0.add(t1); //t0.norm();//33
        FP_FP256BN_sub(&t0,&t0,&t2);			//t0.sub(t2);
        FP_FP256BN_norm(&t0);					//t0.norm();//34
        //FP_FP256BN_copy(&t1,&t4);				//t1.copy(t4);
        FP_FP256BN_mul(&t1,&t4,&y3);			//t1.mul(y3);//35
        //FP_FP256BN_copy(&t2,&t0);				//t2.copy(t0);
        FP_FP256BN_mul(&t2,&t0,&y3);			//t2.mul(y3);//36
        //FP_FP256BN_copy(&y3,&x3);				//y3.copy(x3);
        FP_FP256BN_mul(&y3,&x3,&z3);			//y3.mul(z3);//37
        FP_FP256BN_add(&(P->y),&y3,&t2);			//y3.add(t2); //y3.norm();//38
        FP_FP256BN_mul(&x3,&x3,&t3);			//x3.mul(t3);//39
        FP_FP256BN_sub(&(P->x),&x3,&t1);			//x3.sub(t1);//40
        FP_FP256BN_mul(&z3,&z3,&t4);			//z3.mul(t4);//41
        //FP_FP256BN_copy(&t1,&t3);				//t1.copy(t3);
        FP_FP256BN_mul(&t1,&t3,&t0);			//t1.mul(t0);//42
        FP_FP256BN_add(&(P->z),&z3,&t1);			//z3.add(t1);
        //FP_FP256BN_copy(&(P->x),&x3);			//x.copy(x3);
        FP_FP256BN_norm(&(P->x));				//x.norm();
        //FP_FP256BN_copy(&(P->y),&y3);			//y.copy(y3);
        FP_FP256BN_norm(&(P->y));				//y.norm();
        //FP_FP256BN_copy(&(P->z),&z3);			//z.copy(z3);
        FP_FP256BN_norm(&(P->z));				//z.norm();
    }
}

static int teq(int32_t b,int32_t c)
{
    int32_t x=b^c;
    x-=1;  // if x=0, x now -1
    return (int)((x>>31)&1);
}

/* Constant time select from pre-computed table */
void ECP_FP256BN_select(ECP_FP256BN *P,ECP_FP256BN W[], int32_t b)
{
    ECP_FP256BN MP;
    int32_t m=b>>31;
    int32_t babs=(b^m)-m;

    babs=(babs-1)/2;

    ECP_FP256BN_cmove(P,&W[0],teq(babs,0));  // conditional move
    ECP_FP256BN_cmove(P,&W[1],teq(babs,1));
    ECP_FP256BN_cmove(P,&W[2],teq(babs,2));
    ECP_FP256BN_cmove(P,&W[3],teq(babs,3));
    ECP_FP256BN_cmove(P,&W[4],teq(babs,4));
    ECP_FP256BN_cmove(P,&W[5],teq(babs,5));
    ECP_FP256BN_cmove(P,&W[6],teq(babs,6));
    ECP_FP256BN_cmove(P,&W[7],teq(babs,7));

    ECP_FP256BN_copy(&MP,P);
    ECP_FP256BN_neg(&MP);  // minus P
    ECP_FP256BN_cmove(P,&MP,(int)(m&1));
}

/* SU=8 */
void ECP_FP256BN_neg(ECP_FP256BN *P)
{
//    if (ECP_FP256BN_isinf(P)) return;
    FP_FP256BN_neg(&(P->y),&(P->y));
    FP_FP256BN_norm(&(P->y));
}

/* Set P-=Q */
/* SU=16 */
void  ECP_FP256BN_sub(ECP_FP256BN *P,ECP_FP256BN *Q)
{
    ECP_FP256BN_neg(Q);
    ECP_FP256BN_add(P,Q);
    ECP_FP256BN_neg(Q);
}

/* Convert P to octet string */
void ECP_FP256BN_toOctet(octet *W,ECP_FP256BN *P)
{
    BIG_256_28 x,y;
    ECP_FP256BN_get(x,y,P);
    W->len=2*MODBYTES_256_28+1;
    W->val[0]=4;
    BIG_256_28_toBytes(&(W->val[1]),x);
    BIG_256_28_toBytes(&(W->val[MODBYTES_256_28+1]),y);
}

/* ************************************************************************** */
/* Public Functions                                                           */
/* ************************************************************************** */

int ecp_FP256BN_deserialize(ECP_FP256BN *point_out,
                        uint8_t *buffer)
{
    // 1) Check that serialized point was properly formatted.
    if (0x4 != buffer[0])
        return -2;

    // 2) Get the x,y coordinates.
    BIG_256_28 wx, wy;
    BIG_256_28_fromBytes(wx, (char*)&(buffer[1]));
    BIG_256_28_fromBytes(wy, (char*)&(buffer[MODBYTES_256_28+1]));

    // 3) Check the coordinates are valid Fp points
    //  (i.e. that they are less-than modulus)
    //  (step 2 in X9.62 Sec 5.2.2)
    BIG_256_28 q;
    BIG_256_28_rcopy(q, Modulus_FP256BN);
    if (1 == BIG_256_28_comp(wx, q))
        return -1;
    if (1 == BIG_256_28_comp(wy, q))
        return -1;

    // 4) Check that point is on curve (ECP_FP256BN_set does this implicitly).
    //      (step 3 in X9.62 Sec 5.2.2)
    if (!ECP_FP256BN_set(point_out, wx, wy)) {
        return -1;
    }

    // 5) Check that point is not the identity.
    //      (step 1 in X9.62 Sec 5.2.2)
    if (ECP_FP256BN_isinf(point_out)) {
        return -1;
    }

    // 6) Check that point is in the proper subgroup
    //  (step 4 in X9.62 Sec 5.2.2)
    //  (order*point == inf is equivalent to cofactor*point != inf).
    ECP_FP256BN point_copy;
    ECP_FP256BN_copy(&point_copy, point_out);

    BIG_256_28 cof;
    BIG_256_28_rcopy(cof, CURVE_Cof_FP256BN);
    ECP_FP256BN_mul(&point_copy, cof);

    if (ECP_FP256BN_isinf(&point_copy)) {
        return -1;
    }

    return 0;
}

void ecp_FP256BN_serialize(uint8_t *buffer_out,
                       ECP_FP256BN *point)
{
    octet as_oct = {.len = 0,
                    .max = ECP_FP256BN_LENGTH,
                    .val = (char*)buffer_out};

    ECP_FP256BN_toOctet(&as_oct, point);
}

int point_on_G1_not_infinity(char * point_buffer, size_t buffer_size, ECP_FP256BN * point)
{
    int ret = 0;
    char x_buffer[MODBYTES_256_28] = {0};
    char y_buffer[MODBYTES_256_28] = {0};
    BIG_256_28 x, y;
    uint8_t new_point_buffer[2*MODBYTES_256_28 + 1] = {0x0};

    ENSURE_OR_GO_EXIT(buffer_size == MODBYTES_256_28*2+1);
    ENSURE_OR_GO_EXIT(point_buffer != NULL);
    ENSURE_OR_GO_EXIT(point != NULL);

    // Get x and y from point
    memcpy(&x_buffer[0], &point_buffer[1], MODBYTES_256_28);
    memcpy(&y_buffer[0], &point_buffer[1+MODBYTES_256_28], MODBYTES_256_28);

    // x mod q, y mod q
    BIG_256_28_fromBytes(x, (char *)x_buffer);
    BIG_256_28_fromBytes(y, (char *)y_buffer);
    BIG_256_28_mod(x, Modulus_FP256BN);
    BIG_256_28_mod(y, Modulus_FP256BN);
    BIG_256_28_toBytes((char *)x_buffer, x);
    BIG_256_28_toBytes((char *)y_buffer, y);

    // new point after mod q
    new_point_buffer[0] = 0x04;
    memcpy(&new_point_buffer[1], &x_buffer[0], MODBYTES_256_28);
    memcpy(&new_point_buffer[1+MODBYTES_256_28], &y_buffer[0], MODBYTES_256_28);
    ret = ecp_FP256BN_deserialize(point, (uint8_t*)new_point_buffer);

    if (ret == 0) {
        // Point Is On G1 and Not Infinity
        return 1;
    }

exit:
    return 0;
}

/* clang-format on */