/* Copyright 2021 NXP
 *
 * SPDX-License-Identifier: Apache-2.0
 * */

/* ************************************************************************** */
/* Includes                                                                   */
/* ************************************************************************** */
#include <stdio.h>
#include <string.h>
#include "nxEnsure.h"
#include <nxLog_App.h>

#include <fsl_sss_se05x_apis.h>
#include "ex_sss_fido_hash.h"
#include "ex_sss_fido_big_256_28.h"
#include "ex_sss_fido_ecp.h"

/* ************************************************************************** */
/* Local Defines                                                              */
/* ************************************************************************** */

/* ************************************************************************** */
/* Structures and Typedefs                                                    */
/* ************************************************************************** */

/* ************************************************************************** */
/* Global Variables                                                           */
/* ************************************************************************** */

/* clang-format off */
static BIG_256_28 big_value_two= {0x2,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0};
/* clang-format on */

extern BIG_256_28 Modulus_FP256BN;
extern BIG_256_28 CURVE_Order_FP256BN;
extern BIG_256_28 CURVE_B_FP256BN;

/* ************************************************************************** */
/* Static function declarations                                               */
/* ************************************************************************** */
static int power(BIG_256_28 output, BIG_256_28 base, BIG_256_28 exponent, BIG_256_28 modulus);
static void gcd(BIG_256_28 output, BIG_256_28 a, BIG_256_28 b);
static int order(BIG_256_28 output, BIG_256_28 p, BIG_256_28 b);
static int convertx2e(BIG_256_28 output, BIG_256_28 x, BIG_256_28 e);
static int STonelli(BIG_256_28 output, BIG_256_28 n, BIG_256_28 p);

/* ************************************************************************** */
/* Private Functions                                                          */
/* ************************************************************************** */

static int power(BIG_256_28 output, BIG_256_28 base, BIG_256_28 exponent, BIG_256_28 modulus)
{
    BIG_256_28 temp_value = {0};
    BIG_256_28 local_base, local_exp, local_mod;
    //    char exp_buffer[MODBYTES_256_28] = {0};

    BIG_256_28_copy(local_base, base);
    BIG_256_28_copy(local_exp, exponent);
    BIG_256_28_copy(local_mod, modulus);

    //    BIG_256_28_toBytes((char *)exp_buffer, local_exp);
    //    LOG_MAU8_I("exp", (unsigned char *)exp_buffer, MODBYTES_256_28);

    BIG_256_28_one(output);
    BIG_256_28_mod(local_base, local_mod);
    while (!BIG_256_28_iszilch(local_exp)) {
        BIG_256_28_copy(temp_value, local_exp);                           // temp_value = exponent;
        BIG_256_28_mod(temp_value, big_value_two);                        // temp_value = temp_value % 2;
        if (BIG_256_28_isunity(temp_value)) {                             // if (exponent % 2 == 1)
            BIG_256_28_modmul(temp_value, output, local_base, local_mod); // temp_value = (result * base)% modulus;
            BIG_256_28_copy(output, temp_value);                          // output = temp_value
        }
        //        BIG_256_28_shr(local_exp, 1);            // exponent = exponent >> 1;
        BIG_256_28_sdiv(local_exp, big_value_two);
        //        BIG_256_28_moddiv(local_exp, local_exp, big_value_two, local_mod);

        //        BIG_256_28_toBytes((char *)exp_buffer, local_exp);
        //        LOG_MAU8_I("exp", (unsigned char *)exp_buffer, MODBYTES_256_28);

        BIG_256_28_modmul(temp_value, local_base, local_base, local_mod); // temp_value = (base * base) % modulus;
        BIG_256_28_copy(local_base, temp_value);                          // base = temp_value
    }

    return 0;
}

//  utility function to find gcd
static void gcd(BIG_256_28 output, BIG_256_28 a, BIG_256_28 b)
{
    BIG_256_28 local_a, local_b;
    BIG_256_28_copy(local_a, a);
    BIG_256_28_copy(local_b, b);

    if (BIG_256_28_iszilch(local_b)) { // if (b == 0)
        BIG_256_28_copy(output, local_a);
        return;
    }
    else {
        BIG_256_28 temp;
        BIG_256_28_copy(temp, local_a); // temp = a;
        BIG_256_28_mod(temp, local_b);  // temp %= b;
        gcd(output, local_b, temp);
        return;
    }
}

//  Returns k such that b^k = 1 (mod p)
static int order(BIG_256_28 output, BIG_256_28 p, BIG_256_28 b)
{
    BIG_256_28 temp;
    BIG_256_28 local_p, local_b;
    BIG_256_28_copy(local_p, p);
    BIG_256_28_copy(local_b, b);

    gcd(temp, local_p, local_b);
    if (!BIG_256_28_isunity(temp)) { // (gcd(p, b) != 1)
        LOG_W("p and b are not co-prime.\n");
        return -1;
    }

    /* clang-format off */
    //  Initializing k with first odd prime number
    BIG_256_28 k = {0x3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0};
    /* clang-format on */
    while (1) {
        power(temp, b, k, p);
        if (BIG_256_28_isunity(temp)) { // (pow(b, k, p) == 1
            BIG_256_28_copy(output, k);
            return 0;
        }
        BIG_256_28_inc(k, 1);
    }
}

//  function return  p - 1 (= x argument) as  x * 2^e,
//  where x will be odd  sending e as reference because
//  updation is needed in actual e
static int convertx2e(BIG_256_28 output, BIG_256_28 x, BIG_256_28 e)
{
    BIG_256_28 temp, local_x;
    BIG_256_28_copy(local_x, x);

    BIG_256_28_zero(e);

    BIG_256_28_copy(temp, local_x);      // temp = x;
    BIG_256_28_mod(temp, big_value_two); // temp = temp % 2;

    while (BIG_256_28_iszilch(temp)) {
        //       BIG_256_28_shr(local_x, 1);           // x /= 2;
        BIG_256_28_sdiv(local_x, big_value_two);
        //        BIG_256_28_moddiv(local_x, big_value_two);
        BIG_256_28_inc(e, 1); // e++;

        BIG_256_28_copy(temp, local_x);      // temp = x;
        BIG_256_28_mod(temp, big_value_two); // temp = temp % 2;
    }

    BIG_256_28_copy(output, local_x); // output = x;
    return 0;
}

//  Main function for finding the modular square root
static int STonelli(BIG_256_28 output, BIG_256_28 n, BIG_256_28 p)
{
    //  a and p should be coprime for finding the modular
    // square root
    BIG_256_28 temp_value1, temp_value2;
    BIG_256_28 local_n, local_p;

    BIG_256_28_copy(local_n, n);
    BIG_256_28_copy(local_p, p);

    gcd(temp_value1, local_n, local_p);
    if (!BIG_256_28_isunity(temp_value1)) {
        LOG_W("a and p are not coprime\n");
        return -1;
    }

    //  If below expression return (p - 1)  then modular
    // square root is not possible
    BIG_256_28_copy(temp_value2, local_p); // temp_value2 = p;
    BIG_256_28_dec(temp_value2, 1);        // temp_value2 = p - 1;
                                           //    BIG_256_28_shr(temp_value2, 1);                 // (p - 1) / 2
    BIG_256_28_sdiv(temp_value2, big_value_two);
    //    BIG_256_28_moddiv(temp_value2, big_value_two);

    power(temp_value1, local_n, temp_value2, local_p);             // pow(n, (p - 1) / 2, p)
    BIG_256_28_copy(temp_value2, local_p);                         // temp_value2 = p;
    BIG_256_28_dec(temp_value2, 1);                                // temp_value2 = p - 1;
    if (memcmp(temp_value1, temp_value2, sizeof(BIG_256_28)) == 0) // if (pow(value, n, (p - 1) / 2, p) == (p - 1))
    {
        LOG_D("no sqrt possible\n");
        return -1;
    }

    //  expressing p - 1, in terms of s * 2^e,  where s
    // is odd number
    BIG_256_28 s, e;
    BIG_256_28_copy(temp_value1, local_p); // temp_value1 = p;
    BIG_256_28_dec(temp_value1, 1);        // temp_value1 = temp_value1 - 1;
    convertx2e(s, temp_value1, e);

    //  finding smallest q such that q ^ ((p - 1) / 2)
    //  (mod p) = p - 1
    /* clang-format off */
    BIG_256_28 q = {0x2,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0};
    /* clang-format on */
    for (;;) {
        // q - 1 is in place of  (-1 % p)

        BIG_256_28_copy(temp_value2, local_p); // temp_value2 = p;
        BIG_256_28_dec(temp_value2, 1);        // temp_value2 = p - 1;
                                               //        BIG_256_28_shr(temp_value2, 1);         // (p - 1) / 2
        BIG_256_28_sdiv(temp_value2, big_value_two);
        //        BIG_256_28_moddiv(temp_value2, big_value_two);
        power(temp_value1, q, temp_value2, local_p);
        BIG_256_28_copy(temp_value2, local_p); // temp_value2 = p;
        BIG_256_28_dec(temp_value2, 1);        // temp_value2 = p - 1;

        if (memcmp(temp_value1, temp_value2, sizeof(BIG_256_28)) == 0) {
            break;
        }

        BIG_256_28_inc(q, 1);
    }

    //  Initializing variable x, b and g
    BIG_256_28 x, b, g, r;

    BIG_256_28_copy(temp_value1, s); // temp_value1 = s;
    BIG_256_28_inc(temp_value1, 1);  // temp_value1 = s + 1;
                                     //    BIG_256_28_shr(temp_value1, 1);         // temp_value1 = (s + 1) / 2
    BIG_256_28_sdiv(temp_value1, big_value_two);
    //    BIG_256_28_moddiv(temp_value1, big_value_two);
    power(x, local_n, temp_value1, local_p); // x = pow(n, (s + 1) / 2, p);
    power(b, local_n, s, local_p);           // b = pow(n, s, p);
    power(g, q, s, local_p);                 // g = pow(q, s, p);
    BIG_256_28_copy(r, e);                   // temp_value1 = s;

    // keep looping until b become 1 or m becomes 0
    while (1) {
        BIG_256_28 m = {0};
        for (; BIG_256_28_comp(m, r) == -1; BIG_256_28_inc(m, 1)) { //m < r; m++
            if (order(temp_value1, p, b) == -1) {
                return -1;
            }

            //  finding m such that b^ (2^m) = 1
            power(temp_value2, big_value_two, m, p);
            if (BIG_256_28_comp(temp_value1, temp_value2) == 0) { // (order(p, b) == pow(2, m)
                break;
            }
        }

        if (BIG_256_28_iszilch(m)) { // if (m == 0)
            BIG_256_28_copy(output, x);
        }

        // updating value of x, g and b according to
        // algorithm
        BIG_256_28_sub(temp_value1, r, m);                 // temp_value1 = r - m;
        BIG_256_28_dec(temp_value1, 1);                    // temp_value1 = r - m - 1;
        power(temp_value2, big_value_two, temp_value1, p); // temp_value2 = pow(2, r - m - 1, p)
        power(temp_value1, g, temp_value2, p);             // temp_value1 = pow(g, pow(2, r - m - 1, p), p)
        BIG_256_28_modmul(temp_value2, x, temp_value1, p); // temp_value2 = (x * pow(g, pow(2, r - m - 1, p), p)) % p
        BIG_256_28_copy(x, temp_value2);                   // x = (x * pow(g, pow(2, r - m - 1), p)) % p;

        BIG_256_28_sub(temp_value1, r, m);                 // temp_value1 = r - m;
        power(temp_value2, big_value_two, temp_value1, p); // temp_value2 = pow(2, r - m, p)
        power(temp_value1, g, temp_value2, p);             // temp_value2 = pow(g, pow(2, r - m, p), p);
        BIG_256_28_copy(g, temp_value1);                   // g = pow(g, pow(2, r - m), p);

        BIG_256_28_modmul(temp_value1, b, g, p); // (b * g) % p;
        BIG_256_28_copy(b, temp_value1);         // b = (b * g) % p;

        if (BIG_256_28_isunity(b)) {
            BIG_256_28_copy(output, x); // return x;
            return 0;
        }

        BIG_256_28_copy(r, m); // r = m;
    }
}

sss_status_t H(BIG_256_28 output_BIG, uint8_t *output_buffer, uint8_t *input_buffer, size_t input_buff_len)
{
    sss_status_t status = kStatus_SSS_Fail;
    BIG_256_28 sha256_BIG;

    ENSURE_OR_GO_EXIT(output_BIG != NULL);
    ENSURE_OR_GO_EXIT(output_buffer != NULL);
    ENSURE_OR_GO_EXIT(input_buffer != NULL);

    // SHA256(input)
    big_256_28_from_hash(&sha256_BIG, input_buffer, input_buff_len);
    //    BIG_256_28_rawoutput(sha256_BIG);

    // H(i|m) = SHA256(input) mod p
    BIG_256_28_mod(sha256_BIG, CURVE_Order_FP256BN);
    //    BIG_256_28_rawoutput(sha256_BIG);

    BIG_256_28_toBytes((char *)output_buffer, sha256_BIG);
    //    LOG_MAU8_I("H output", (unsigned char *)output_buffer, MODBYTES_256_28);

    BIG_256_28_copy(output_BIG, sha256_BIG);

    status = kStatus_SSS_Success;

exit:
    return status;
}

/* ************************************************************************** */
/* Public Functions                                                           */
/* ************************************************************************** */

sss_status_t HG1(BIG_256_28 m, BIG_256_28 x, BIG_256_28 y, int *sc_index, BIG_256_28 yc)
{
    int ret;
    sss_status_t status = kStatus_SSS_Fail;
    uint32_t i          = 0;
    BIG_256_28 local_m;
    uint8_t m_buffer[MODBYTES_256_28]      = {0};
    char pre_x_buffer[4 + MODBYTES_256_28] = {0}; // i|m
    BIG_256_28 x_BIG, z_BIG, y_BIG;
    uint8_t x_buffer[MODBYTES_256_28] = {0};
    uint8_t z_buffer[MODBYTES_256_28] = {0};
    uint8_t y_buffer[MODBYTES_256_28] = {0};
    BIG_256_28 q_y_Big;
    uint8_t q_y_buffer[MODBYTES_256_28]       = {0};
    uint8_t p_buffer[MODBYTES_256_28 * 2 + 1] = {0};
    ECP_FP256BN p_ECPoint;

    ENSURE_OR_GO_EXIT(m != NULL);
    ENSURE_OR_GO_EXIT(x != NULL);
    ENSURE_OR_GO_EXIT(y != NULL);
    ENSURE_OR_GO_EXIT(sc_index != NULL);
    ENSURE_OR_GO_EXIT(yc != NULL);

    BIG_256_28_copy(local_m, m);
    BIG_256_28_toBytes((char *)m_buffer, local_m);

    for (i = 0; i < 232; i++) {
        pre_x_buffer[0] = (uint8_t)((i >> 3 * 8) & 0xFF);
        pre_x_buffer[1] = (uint8_t)((i >> 2 * 8) & 0xFF);
        pre_x_buffer[2] = (uint8_t)((i >> 1 * 8) & 0xFF);
        pre_x_buffer[3] = (uint8_t)((i >> 0 * 8) & 0xFF);
        memcpy(&pre_x_buffer[4], m_buffer, sizeof(m_buffer));

        // x = H(i|m) = SHA256(i|m) mod p
        status = H(x_BIG, x_buffer, (uint8_t *)pre_x_buffer, sizeof(pre_x_buffer));
        ENSURE_OR_GO_EXIT(status == kStatus_SSS_Success);
        LOG_MAU8_D("x", (unsigned char *)x_buffer, MODBYTES_256_28);

        // z = x^3 + b mod q
        BIG_256_28_modmul(z_BIG, x_BIG, x_BIG, Modulus_FP256BN);
        BIG_256_28_modmul(z_BIG, z_BIG, x_BIG, Modulus_FP256BN);
        BIG_256_28_add(z_BIG, z_BIG, CURVE_B_FP256BN);
        BIG_256_28_mod(z_BIG, Modulus_FP256BN);
        BIG_256_28_toBytes((char *)z_buffer, z_BIG);
        LOG_MAU8_D("z", (unsigned char *)z_buffer, MODBYTES_256_28);

        // y = sqrt(z) mod q.
        ret = STonelli(y_BIG, z_BIG, Modulus_FP256BN);
        if (ret != -1) {
            BIG_256_28_toBytes((char *)y_buffer, y_BIG); // y
            LOG_MAU8_D("sqrt(z)", (unsigned char *)y_buffer, MODBYTES_256_28);

            BIG_256_28_sub(q_y_Big, Modulus_FP256BN, y_BIG); // q - y
            BIG_256_28_toBytes((char *)q_y_buffer, q_y_Big);
            LOG_MAU8_D("q - y", (unsigned char *)q_y_buffer, MODBYTES_256_28);
            if (BIG_256_28_comp(y_BIG, q_y_Big) > 0) { // Set y = min(y, q - y).
                BIG_256_28_copy(y_BIG, q_y_Big);
                BIG_256_28_toBytes((char *)y_buffer, y_BIG);
                LOG_MAU8_D("y", (unsigned char *)y_buffer, MODBYTES_256_28);
            }

            // ECPoint p(x, y)
            p_buffer[0] = 0x04;
            memcpy(&p_buffer[1], x_buffer, MODBYTES_256_28);
            memcpy(&p_buffer[1 + MODBYTES_256_28], y_buffer, MODBYTES_256_28);
            ret = point_on_G1_not_infinity((char *)p_buffer, sizeof(p_buffer), &p_ECPoint);

            if (ret == 1) {
                BIG_256_28_copy(x, x_BIG);
                BIG_256_28_copy(y, y_BIG);
                *sc_index = i;
                BIG_256_28_copy(yc, y_BIG);
                break;
            }
        }
    }

    status = kStatus_SSS_Success;
exit:
    return status;
}
