/*
 *
 * Copyright 2018,2021 NXP
 * SPDX-License-Identifier: Apache-2.0
 */

/* ************************************************************************** */
/* Includes                                                                   */
/* ************************************************************************** */

#include <ex_sss.h>
#include <ex_sss_boot.h>
#include <fsl_sss_se05x_apis.h>
#include <nxEnsure.h>
#include <nxLog_App.h>

#include "ex_sss_fido_big_256_28.h"
#include "ex_sss_fido_ecp.h"
#include "ex_sss_fido_sqrt.h"
#if SSS_HAVE_APPLET_SE05X_IOT
#include "se05x_APDU_apis.h"
#include <fsl_sss_se05x_policy.h>
#endif

/* ************************************************************************** */
/* Local Defines                                                              */
/* ************************************************************************** */
#define KEY_BIT_LENGTH 256
/* ************************************************************************** */
/* Structures and Typedefs                                                    */
/* ************************************************************************** */

/* ************************************************************************** */
/* Global Variables                                                           */
/* ************************************************************************** */

static ex_sss_boot_ctx_t gex_sss_ecdaa_boot_ctx;

/* clang-format off */
BIG_256_28 CURVE_Order_FP256BN= {0x10B500D,0x2D536CD,0x9921AF6,0x65FB129,0x49E0CDC,0x5EEE71A,0xD46E5F2,0xFFFCF0C,0xFFFFFFF,0xF};
BIG_256_28 CURVE_Gx_FP256BN= {0x1,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0};
BIG_256_28 CURVE_Gy_FP256BN= {0x2,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0};
/* clang-format on */

/* ************************************************************************** */
/* Static function declarations                                               */
/* ************************************************************************** */

/* ************************************************************************** */
/* Private Functions                                                          */
/* ************************************************************************** */

/* ************************************************************************** */
/* Public Functions                                                           */
/* ************************************************************************** */

#define EX_SSS_BOOT_PCONTEXT (&gex_sss_ecdaa_boot_ctx)
#define EX_SSS_BOOT_DO_ERASE 1
#define EX_SSS_BOOT_EXPOSE_ARGC_ARGV 0
#define SSS_SE05X_RESID_ECDAA_RANDOM_KEY_START 0x7DB00000
#define SSS_SE05X_RESID_ECDAA_RANDOM_KEY_NUMBER 0x00010000

#include <ex_sss_main_inc.h>

/* Reserved Random key id is 0x7DB0,0000 -> 0x7DB0,FFFF
*  0x7DB0,0000 is used for None auth session and platform scp03 session
*  Other id = (auth_id % 0x0000,FFFF) + 0x7DB0,0001
*/
static sss_status_t ex_fido_get_ecdaa_random_key_id(Se05xSession_t *context, uint32_t *random_keyId)
{
    sss_status_t retval = kStatus_SSS_Fail;

    ENSURE_OR_GO_EXIT(context != NULL);
    ENSURE_OR_GO_EXIT(random_keyId != NULL);

    if (context->hasSession == 1) {
        *random_keyId = context->auth_id % (SSS_SE05X_RESID_ECDAA_RANDOM_KEY_NUMBER - 1) +
                        SSS_SE05X_RESID_ECDAA_RANDOM_KEY_START + 1;
    }
    else {
        *random_keyId = SSS_SE05X_RESID_ECDAA_RANDOM_KEY_START;
    }

    retval = kStatus_SSS_Success;
exit:
    return retval;
}

static sss_status_t ex_sss_initialise_attst_data(sss_se05x_attst_data_t *attst_data)
{
    sss_status_t status = kStatus_SSS_Fail;
    size_t i;

    if (attst_data != NULL) {
        memset(attst_data, 0, sizeof(sss_se05x_attst_data_t));

        for (i = 0; i < SE05X_MAX_ATTST_DATA; i++) {
#if SSS_HAVE_SE05X_VER_GTE_07_02
            attst_data->data[i].cmdLen     = sizeof(attst_data->data[i].cmd);
            attst_data->data[i].objSizeLen = sizeof(attst_data->data[i].objSize);
#else
            attst_data->data[i].outrandomLen = sizeof(attst_data->data[i].outrandom);
#endif //#if SSS_HAVE_SE05X_VER_GTE_07_02
            attst_data->data[i].timeStampLen = sizeof(attst_data->data[i].timeStamp);
            attst_data->data[i].chipIdLen    = sizeof(attst_data->data[i].chipId);
            attst_data->data[i].attributeLen = sizeof(attst_data->data[i].attribute);
            attst_data->data[i].signatureLen = sizeof(attst_data->data[i].signature);
        }
        attst_data->valid_number = SE05X_MAX_ATTST_DATA;

        status = kStatus_SSS_Success;
    }
    return status;
}

sss_status_t ex_sss_entry(ex_sss_boot_ctx_t *pCtx)
{
    sss_status_t status        = kStatus_SSS_Fail;
    sss_asymmetric_t ctx_asymm = {0};
    sss_rng_context_t ctx_rng  = {0};
    size_t keylen              = KEY_BIT_LENGTH / 8;
    uint8_t signature[256]     = {0};
    size_t signatureLen;
    BIG_256_28 host_Bx                                                  = {0};
    uint8_t sc_buffer[MODBYTES_256_28 + 4]                              = {0};
    uint32_t sc_buffer_len                                              = sizeof(sc_buffer); // BigNumberToB(i,4) | m
    char host_Bx_buffer[MODBYTES_256_28]                                = {0};
    char yc_buffer[MODBYTES_256_28]                                     = {0};
    char host_B_buffer[MODBYTES_256_28 * 2 + 1]                         = {0}; // 04 | H(sc) | yc
    char q_buffer[MODBYTES_256_28 * 2 + 1]                              = {0};
    size_t q_key_byte_len                                               = sizeof(q_buffer);
    size_t q_key_bit_len                                                = sizeof(q_buffer) * 8;
    char u1_buffer[MODBYTES_256_28 * 2 + 1]                             = {0};
    size_t u1_key_bit_len                                               = sizeof(u1_buffer) * 8;
    size_t u1_key_byte_len                                              = sizeof(u1_buffer);
    char pre_c2_buffer[(MODBYTES_256_28 * 2 + 1) * 3 + MODBYTES_256_28] = {0}; // U1|P1|Q|m
    BIG_256_28 c2                                                       = {0};
    char c2_buffer[MODBYTES_256_28]                                     = {0};
    uint8_t m_buffer[MODBYTES_256_28]                                   = {0};
    BIG_256_28 c1                                                       = {0};
    char c1_buffer[MODBYTES_256_28]                                     = {0};
    char pre_c1_buffer[MODBYTES_256_28 * 2]                             = {0}; // n|c_2
    BIG_256_28 s1                                                       = {0};
    char s1_buffer[MODBYTES_256_28]                                     = {0};
    ECP_FP256BN B_s_1;
    char B_s_1_buffer[MODBYTES_256_28 * 2 + 1] = {0};
    ECP_FP256BN Q_c_1;
    char Q_c_1_buffer[MODBYTES_256_28 * 2 + 1] = {0};
    ECP_FP256BN U_verify;
    char U_verify_buffer[MODBYTES_256_28 * 2 + 1]                               = {0};
    char pre_c_2_verify_buffer[(MODBYTES_256_28 * 2 + 1) * 3 + MODBYTES_256_28] = {0}; //U_verify|P1|Q|m
    BIG_256_28 c_2_verify                                                       = {0};
    char c_2_verify_buffer[MODBYTES_256_28]                                     = {0};
    char P1_256_buffer[MODBYTES_256_28 * 2 + 1]                                 = {0};
    char pre_c_1_verify_buffer[MODBYTES_256_28 * 2]                             = {0}; //n|c_2_verify
    BIG_256_28 c_1_verify                                                       = {0};
    char c_1_verify_buffer[MODBYTES_256_28]                                     = {0};
    ECP_FP256BN A;
    char A_buffer[MODBYTES_256_28 * 2 + 1]                                     = {0};
    char C_buffer[MODBYTES_256_28 * 2 + 1]                                     = {0};
    char D_buffer[MODBYTES_256_28 * 2 + 1]                                     = {0};
    char S_buffer[MODBYTES_256_28 * 2 + 1]                                     = {0};
    char l_buffer[MODBYTES_256_28]                                             = {0};
    char R_buffer[MODBYTES_256_28 * 2 + 1]                                     = {0};
    size_t R_buffer_len                                                        = sizeof(R_buffer);
    size_t s_key_bit_len                                                       = sizeof(S_buffer) * 8;
    size_t s_key_byte_len                                                      = sizeof(S_buffer);
    char T_buffer[MODBYTES_256_28 * 2 + 1]                                     = {0};
    size_t T_buffer_len                                                        = sizeof(T_buffer);
    char W_buffer[MODBYTES_256_28 * 2 + 1]                                     = {0};
    size_t W_buffer_len                                                        = sizeof(W_buffer);
    char U_buffer[MODBYTES_256_28 * 2 + 1]                                     = {0};
    size_t U_buffer_len                                                        = sizeof(U_buffer);
    BIG_256_28 hash_KRD_BIG                                                    = {0};
    char hash_KRD_buffer[MODBYTES_256_28]                                      = {0};
    BIG_256_28 c2_1_BIG                                                        = {0};
    char c2_1_buffer[MODBYTES_256_28]                                          = {0};
    char pre_c2_1_buffer[(MODBYTES_256_28 * 2 + 1) * 3 + 25 + MODBYTES_256_28] = {0}; // U|S|W|AppID|H(KRD)
    BIG_256_28 c_BIG                                                           = {0};
    char c_buffer[MODBYTES_256_28]                                             = {0};
    char pre_c_buffer[MODBYTES_256_28 * 2]                                     = {0}; // n|c_2
    BIG_256_28 s_BIG                                                           = {0};
    char s_buffer[MODBYTES_256_28]                                             = {0};
    ECP_FP256BN S_s;
    char S_s_buffer[MODBYTES_256_28 * 2 + 1] = {0};
    ECP_FP256BN W_c;
    char W_c_buffer[MODBYTES_256_28 * 2 + 1]                                           = {0};
    char pre_c_2_2_verify_buffer[(MODBYTES_256_28 * 2 + 1) * 3 + 25 + MODBYTES_256_28] = {
        0}; // U_verify|S|W|AppID|H(KRD)
    BIG_256_28 c_2_2_verify_BIG                   = {0};
    char c_2_2_verify_buffer[MODBYTES_256_28]     = {0};
    char pre_c_verify_buffer[MODBYTES_256_28 * 2] = {0}; // n|c_2_verify
    BIG_256_28 c_verify_BIG                       = {0};
    char c_verify_buffer[MODBYTES_256_28]         = {0};
    BIG_256_28 l_BIG                              = {0};

    sss_object_t keyPair_sk, pk_host_b, pk_Q, attst_key, random_key, pk_U1, pk_A, pk_C, pk_D, pk_S, sk_l;
    sss_se05x_session_t *se05x_session;
    smStatus_t sm_status;
    uint8_t random[16] = {0};
    sss_se05x_attst_data_t pk_Q_attst_data;
    uint32_t random_keyId;
    uint8_t n_buffer[32]                  = {0};
    BIG_256_28 n                          = {0};
    BIG_256_28 isk_x                      = {0};
    BIG_256_28 isk_y                      = {0};
    BIG_256_28 isk_y_modInv               = {0};
    uint8_t isk_x_buffer[MODBYTES_256_28] = {0};
    uint8_t isk_y_buffer[MODBYTES_256_28] = {0};
    char appID[]                          = "https://www.nxp.com/se050";
    char keyRegistrationData[100];

    /* Policies for attestation key */
    const sss_policy_u key_withPol = {.type = KPolicy_Asym_Key,
        .auth_obj_id                        = 0, // All users
        .policy                             = {.asymmkey = {
                       .can_Sign              = 0,
                       .can_Verify            = 0,
                       .can_Encrypt           = 0,
                       .can_Decrypt           = 0,
                       .can_Import_Export     = 0,
                       .forbid_Derived_Output = 0,
                       .can_Gen               = 0,
                       .can_KA                = 0,
                       .can_Attest            = 1,
                   }}};

    /* Common rules */
    const sss_policy_u common = {.type = KPolicy_Common,
        .auth_obj_id                   = 0, // All users
        .policy                        = {.common = {
                       .forbid_All  = 0,
                       .can_Read    = 1,
                       .can_Write   = 0,
                       .can_Delete  = 1,
                       .req_Sm      = 0,
                       .req_pcr_val = 0,
                   }}};

    /* create policy set */
    sss_policy_t policy_for_ec_key = {.nPolicies = 2, .policies = {&key_withPol, &common}};
    BIG_256_28 m;
    BIG_256_28 yc = {0};

    int ret = 0;
    ECP_FP256BN B, Q, C, D, R, S, T, W;
    uint8_t B_buffer[2 * MODBYTES_256_28 + 1] = {0};
    BIG_256_28 B_x_BIG, B_y_BIG;
    int sc_index;

    // The authenticator asks the ECDAA Issuer for the B value of the credential.
    // The ECDAA Issuer chooses a nonce BigInteger m.
    // The ECDAA Issuer computes the B value of the credential and sends (sc,yc) to the authenticator.
    LOG_I("Running FIDO Example se05x_FIDO_ECDAA.c");

    LOG_I("EcdaaJoin 1) Get B from Issuer");
    LOG_I("EcdaaJoin 1.1) The ECDAA Issuer chooses a nonce BigInteger m");

    // m is 32B random
    status = sss_rng_context_init(&ctx_rng, &pCtx->session /* Session */);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    status = sss_rng_get_random(&ctx_rng, m_buffer, sizeof(m_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    sss_rng_context_free(&ctx_rng);
    LOG_MAU8_I("m", (unsigned char *)m_buffer, sizeof(m_buffer));

    BIG_256_28_fromBytes(m, (char *)m_buffer);

    LOG_I(
        "EcdaaJoin 1.2) The ECDAA Issuer computes the B value of the credential as B=HG1(m) and send "
        "(sc,yc)=HG1_pre(m) to the authenticator");
    status = HG1(m, B_x_BIG, B_y_BIG, &sc_index, yc);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    B_buffer[0] = 0x04;
    BIG_256_28_toBytes((char *)(&B_buffer[1]), B_x_BIG);
    BIG_256_28_toBytes((char *)(&B_buffer[1 + MODBYTES_256_28]), B_y_BIG);
    ret = ecp_FP256BN_deserialize(&B, B_buffer); // ECPoint B
    ENSURE_OR_GO_CLEANUP(ret == 0);

    LOG_MAU8_I("B", (unsigned char *)B_buffer, sizeof(B_buffer));
    LOG_I("sc index 0x%08x", sc_index);

    LOG_I(
        "EcdaaJoin 2) The authenticator generate secret key (sk) on SE, as a key pair on the BN curve (we only need "
        "the private key part).");

    ENSURE_OR_GO_CLEANUP(pCtx != NULL);

    se05x_session = (sss_se05x_session_t *)(&pCtx->session);

    // Generate sk
    status = sss_key_object_init(&keyPair_sk, &pCtx->ks);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_allocate_handle(&keyPair_sk,
        MAKE_TEST_ID(__LINE__),
        kSSS_KeyPart_Pair,
        kSSS_CipherType_EC_BARRETO_NAEHRIG,
        keylen,
        kKeyObject_Mode_Persistent);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_store_generate_key(&pCtx->ks, &keyPair_sk, KEY_BIT_LENGTH, 0);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    LOG_I(
        "EcdaaJoin 3) The authenticator re-calculates B from the data received from backed (and stores on SE). "
        "Re-calculated from sc and yc as described in FIDO spec.");

    sc_buffer[0] = (uint8_t)((sc_index >> 3 * 8) & 0xFF);
    sc_buffer[1] = (uint8_t)((sc_index >> 2 * 8) & 0xFF);
    sc_buffer[2] = (uint8_t)((sc_index >> 1 * 8) & 0xFF);
    sc_buffer[3] = (uint8_t)((sc_index >> 0 * 8) & 0xFF);
    BIG_256_28_toBytes((char *)(&sc_buffer[4]), m);

    // H(sc) = SHA256(sc) mod p
    status = H(host_Bx, (uint8_t *)host_Bx_buffer, (uint8_t *)sc_buffer, sc_buffer_len);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_D("H(sc)", (unsigned char *)host_Bx_buffer, sizeof(host_Bx_buffer));

    // yc
    BIG_256_28_toBytes(yc_buffer, yc);
    LOG_MAU8_D("yc", (unsigned char *)yc_buffer, sizeof(yc_buffer));

    // Write Host_B public key. Host_B = (H(sc), yc)
    host_B_buffer[0] = 0x04;
    memcpy(&host_B_buffer[1], host_Bx_buffer, sizeof(host_Bx_buffer));
    memcpy(&host_B_buffer[1 + sizeof(host_Bx_buffer)], yc_buffer, sizeof(yc_buffer));
    LOG_MAU8_I("Authenticator B", (unsigned char *)host_B_buffer, sizeof(host_B_buffer));

    status = sss_key_object_init(&pk_host_b, &pCtx->ks);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_allocate_handle(&pk_host_b,
        MAKE_TEST_ID(__LINE__),
        kSSS_KeyPart_Public,
        kSSS_CipherType_EC_BARRETO_NAEHRIG,
        keylen,
        kKeyObject_Mode_Persistent);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_store_set_key(
        &pCtx->ks, &pk_host_b, (uint8_t *)host_B_buffer, sizeof(host_B_buffer), KEY_BIT_LENGTH, NULL, 0);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    LOG_I(
        "EcdaaJoin 4) The authenticator calculation of Q=ECPointMultiply(sk, B), the ECDAA public key (and attested "
        "readback)");
    // Create ECDAA public key Q
    status = sss_key_object_init(&pk_Q, &pCtx->ks);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_allocate_handle(&pk_Q,
        MAKE_TEST_ID(__LINE__),
        kSSS_KeyPart_Public,
        kSSS_CipherType_EC_BARRETO_NAEHRIG,
        keylen,
        kKeyObject_Mode_Persistent);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    q_buffer[0] = 0x04;
    status = sss_key_store_set_key(&pCtx->ks, &pk_Q, (uint8_t *)q_buffer, sizeof(q_buffer), KEY_BIT_LENGTH, NULL, 0);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    // Q = ECPointMultiply(sk, B)
    sm_status = Se05x_API_ECPointMultiply_InputObj(&se05x_session->s_ctx,
        keyPair_sk.keyId,
        pk_host_b.keyId,
        pk_Q.keyId,
        NULL,
        NULL,
        kSE05x_ECPMAlgo_SVDP_DH_PLAIN_XY);
    ENSURE_OR_GO_CLEANUP(sm_status == SM_OK);

    LOG_I(
        "Read Q attested.This is required for FIDO remote provisioning usecase, as opposed to in factory "
        "provisioningThe attestation signature over Q would be used as proof that the sk is indeed on the SE.As FIDO "
        "specifies no anonymity is required in this one time trust establishing operation.");

    // Attestation key NIST-P 256
    status = sss_key_object_init(&attst_key, &pCtx->ks);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_allocate_handle(&attst_key,
        MAKE_TEST_ID(__LINE__),
        kSSS_KeyPart_Pair,
        kSSS_CipherType_EC_NIST_P,
        keylen,
        kKeyObject_Mode_Persistent);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_store_generate_key(&pCtx->ks, &attst_key, KEY_BIT_LENGTH, &policy_for_ec_key);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    // Attestation 16-byte random
    status = sss_rng_context_init(&ctx_rng, &pCtx->session);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    status = sss_rng_get_random(&ctx_rng, random, sizeof(random));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    sss_rng_context_free(&ctx_rng);
    LOG_MAU8_I("Attestation random", (unsigned char *)random, sizeof(random));

    status = ex_sss_initialise_attst_data(&pk_Q_attst_data);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    // Read public key Q with attestation
    status = sss_se05x_key_store_get_key_attst((sss_se05x_key_store_t *)(&pCtx->ks),
        (sss_se05x_object_t *)(&pk_Q),
        (uint8_t *)q_buffer,
        &q_key_byte_len,
        &q_key_bit_len,
        (sss_se05x_object_t *)(&attst_key),
        kAlgorithm_SSS_ECDSA_SHA256,
        random,
        sizeof(random),
        &pk_Q_attst_data);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("Q=B^sk", (unsigned char *)q_buffer, sizeof(q_buffer));

    LOG_I("EcdaaJoin 5) The authenticator generate random (r) on SE, as a private key on the BN curve.");
    // Generate random (r)
    // r has been generatad internally when generating sk. Only get handle here.
    status = ex_fido_get_ecdaa_random_key_id(&se05x_session->s_ctx, &random_keyId);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_init(&random_key, &pCtx->ks);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_get_handle(&random_key, random_keyId);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    LOG_I("EcdaaJoin 6) The authenticator calculation of U1 = ECPointMultiply(r, B) on SE");
    // Create ECDAA public key U1
    status = sss_key_object_init(&pk_U1, &pCtx->ks);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_allocate_handle(&pk_U1,
        MAKE_TEST_ID(__LINE__),
        kSSS_KeyPart_Public,
        kSSS_CipherType_EC_BARRETO_NAEHRIG,
        keylen,
        kKeyObject_Mode_Persistent);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    u1_buffer[0] = 0x04;
    status = sss_key_store_set_key(&pCtx->ks, &pk_U1, (uint8_t *)u1_buffer, sizeof(u1_buffer), KEY_BIT_LENGTH, NULL, 0);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    // U1 = ECPointMultiply(r, B)
    sm_status = Se05x_API_ECPointMultiply_InputObj(&se05x_session->s_ctx,
        random_key.keyId,
        pk_host_b.keyId,
        pk_U1.keyId,
        NULL,
        NULL,
        kSE05x_ECPMAlgo_SVDP_DH_PLAIN_XY);
    ENSURE_OR_GO_CLEANUP(sm_status == SM_OK);

    // Read U1
    status = sss_key_store_get_key(&pCtx->ks, &pk_U1, (uint8_t *)u1_buffer, &u1_key_byte_len, &u1_key_bit_len);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("U1=B^r1", (unsigned char *)u1_buffer, sizeof(u1_buffer));

    LOG_I("EcdaaJoin 7) The authenticator calculation of the Hash (c_1) on the Host");
    //c_2 = H(U1|P1|Q|m)

    // U1|P1|Q|m
    memcpy(&pre_c2_buffer[0], u1_buffer, sizeof(u1_buffer));
    pre_c2_buffer[sizeof(u1_buffer)]      = 0x04;
    pre_c2_buffer[sizeof(u1_buffer) + 32] = 0x01; // ECPoint P1.getAffineX = 1
    pre_c2_buffer[sizeof(u1_buffer) + 64] = 0x02; // ECPoint P1.getAffineY = 2
    memcpy(&pre_c2_buffer[sizeof(u1_buffer) * 2], q_buffer, sizeof(q_buffer));
    memcpy(&pre_c2_buffer[sizeof(u1_buffer) * 3], m_buffer, sizeof(m_buffer));

    // c_2 = H(U1|P1|Q|m) = SHA256(U1|P1|Q|m) mod p
    status = H(c2, (uint8_t *)c2_buffer, (uint8_t *)pre_c2_buffer, sizeof(pre_c2_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("c_2=H(U1|P1|Q|m)", (unsigned char *)c2_buffer, sizeof(c2_buffer));

    //c_1 = H(n|c_2)
    status = sss_rng_context_init(&ctx_rng, &pCtx->session /* Session */);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    // n is 32B random mod p
    status = sss_rng_get_random(&ctx_rng, n_buffer, sizeof(n_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    sss_rng_context_free(&ctx_rng);
    LOG_MAU8_I("n", (unsigned char *)n_buffer, sizeof(n_buffer));

    BIG_256_28_fromBytes(n, (char *)n_buffer);
    BIG_256_28_mod(n, CURVE_Order_FP256BN);
    BIG_256_28_toBytes((char *)n_buffer, n);

    // n|c_2
    memcpy(&pre_c1_buffer[0], n_buffer, sizeof(n_buffer));
    memcpy(&pre_c1_buffer[sizeof(n_buffer)], c2_buffer, sizeof(c2_buffer));

    //c_1 = H(n|c_2) = SHA256(n|c_2) mod p
    status = H(c1, (uint8_t *)c1_buffer, (uint8_t *)pre_c1_buffer, sizeof(pre_c1_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("c_1 = H(n|c_2)", (unsigned char *)c1_buffer, sizeof(c1_buffer));

    LOG_I(
        "EcdaaJoin 8) The authenticator calculation of the ECDAA signature s1 on the SE and send Q, c1, s1, n to "
        "Issuer.");

    /* Asymmetric sign */
    status = sss_asymmetric_context_init(&ctx_asymm, &pCtx->session, &keyPair_sk, kAlgorithm_SSS_ECDAA, kMode_SSS_Sign);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    signatureLen = sizeof(signature);
    /* Do Signing */
    LOG_I("Do Signing");
    LOG_MAU8_I("digest", (uint8_t *)c1_buffer, sizeof(c1_buffer));
    status = sss_asymmetric_sign_digest(&ctx_asymm, (uint8_t *)c1_buffer, sizeof(c1_buffer), signature, &signatureLen);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("signature", signature, signatureLen);
    LOG_I("Signing Successful !!!");
    sss_asymmetric_context_free(&ctx_asymm);

    ENSURE_OR_GO_CLEANUP(signatureLen >= sizeof(s1_buffer));
    // Remove asn.1 header from signature.
    memcpy(&s1_buffer[0], &signature[signatureLen - sizeof(s1_buffer)], sizeof(s1_buffer));
    BIG_256_28_fromBytes(s1, (char *)s1_buffer);

    LOG_I("EcdaaJoin 9.1) Verify on the ECDAA Issuer Backend.");

    ret = point_on_G1_not_infinity(q_buffer, sizeof(q_buffer), &Q);
    ENSURE_OR_GO_CLEANUP(ret == 1); // Point Q Is On G1 and Not Infinity

    ECP_FP256BN_copy(&B_s_1, &B); // B_s_1 = B
    ECP_FP256BN_mul(&B_s_1, s1);  // B_s_1 = B_s_1 * s1
    ecp_FP256BN_serialize((uint8_t *)B_s_1_buffer, &B_s_1);
    LOG_MAU8_I("B^s_1", (uint8_t *)B_s_1_buffer, sizeof(B_s_1_buffer));
    ECP_FP256BN_copy(&Q_c_1, &Q); // Q_c_1 = Q
    ECP_FP256BN_mul(&Q_c_1, c1);  // Q_c_1 = Q_c_1 * c1
    ecp_FP256BN_serialize((uint8_t *)Q_c_1_buffer, &Q_c_1);
    LOG_MAU8_I("Q^c1", (uint8_t *)Q_c_1_buffer, sizeof(Q_c_1_buffer));

    ECP_FP256BN_copy(&U_verify, &B_s_1); // U_verify = B_s_1
    ECP_FP256BN_sub(&U_verify, &Q_c_1);  // U_verify = B_s_1 - Q_c_1
    ecp_FP256BN_serialize((uint8_t *)U_verify_buffer, &U_verify);
    ret = point_on_G1_not_infinity(U_verify_buffer, sizeof(U_verify_buffer), &U_verify);
    ENSURE_OR_GO_CLEANUP(ret == 1); // Point U_verify Is On G1 and Not Infinity
    LOG_MAU8_I("U_verify=B^s1-Q^c1", (uint8_t *)Q_c_1_buffer, sizeof(Q_c_1_buffer));

    // U_verify|P1|Q|m
    memcpy(&pre_c_2_verify_buffer[0], U_verify_buffer, sizeof(U_verify_buffer));
    P1_256_buffer[0] = 0x04;
    BIG_256_28_toBytes((char *)(&P1_256_buffer[1]), CURVE_Gx_FP256BN);
    BIG_256_28_toBytes((char *)(&P1_256_buffer[1 + MODBYTES_256_28]), CURVE_Gy_FP256BN);
    memcpy(&pre_c_2_verify_buffer[sizeof(U_verify_buffer)], P1_256_buffer, sizeof(P1_256_buffer));
    memcpy(&pre_c_2_verify_buffer[sizeof(U_verify_buffer) + sizeof(P1_256_buffer)], q_buffer, sizeof(q_buffer));
    memcpy(&pre_c_2_verify_buffer[sizeof(U_verify_buffer) + sizeof(P1_256_buffer) + sizeof(q_buffer)],
        m_buffer,
        sizeof(m_buffer));

    //c_2_verify = H(U_verify|P1|Q|m) = SHA256(U_verify|P1|Q|m) mod p
    status =
        H(c_2_verify, (uint8_t *)c_2_verify_buffer, (uint8_t *)pre_c_2_verify_buffer, sizeof(pre_c_2_verify_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("c_2_verify = H(U_verify|P1|Q|m)", (unsigned char *)c_2_verify_buffer, sizeof(c_2_verify_buffer));

    // n|c_2_verify
    memcpy(&pre_c_1_verify_buffer[0], n_buffer, sizeof(n_buffer));
    memcpy(&pre_c_1_verify_buffer[sizeof(n_buffer)], c_2_verify_buffer, sizeof(c_2_verify_buffer));

    //c_1_verify = H(n|c_2_verify) = SHA256(n|c_2_verify) mod p
    status =
        H(c_1_verify, (uint8_t *)c_1_verify_buffer, (uint8_t *)pre_c_1_verify_buffer, sizeof(pre_c_1_verify_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("c_1_verify = H(n|c_2_verify)", (unsigned char *)c_1_verify_buffer, sizeof(c_1_verify_buffer));
    LOG_MAU8_I("c_1", (unsigned char *)c1_buffer, sizeof(c1_buffer));

    ENSURE_OR_GO_CLEANUP(memcmp(c_1_verify_buffer, c1_buffer, sizeof(c1_buffer)) == 0);
    LOG_I("Verified. H(n|c_2_verify) == c_1");

    LOG_I("EcdaaJoin 9.2) The ECDAA Issuer create a credential for the SE.");

    // Issuer public key: isk_x and isk_y is 32B random mod p
    status = sss_rng_context_init(&ctx_rng, &pCtx->session /* Session */);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    status = sss_rng_get_random(&ctx_rng, isk_x_buffer, sizeof(isk_x_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    status = sss_rng_get_random(&ctx_rng, isk_y_buffer, sizeof(isk_y_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    sss_rng_context_free(&ctx_rng);

    BIG_256_28_fromBytes(isk_x, (char *)isk_x_buffer);
    BIG_256_28_mod(isk_x, CURVE_Order_FP256BN);
    BIG_256_28_toBytes((char *)isk_x_buffer, isk_x);
    LOG_MAU8_I("isk.x", (unsigned char *)isk_x_buffer, sizeof(isk_x_buffer));

    BIG_256_28_fromBytes(isk_y, (char *)isk_y_buffer);
    BIG_256_28_mod(isk_y, CURVE_Order_FP256BN);
    BIG_256_28_toBytes((char *)isk_y_buffer, isk_y);
    LOG_MAU8_I("isk.y", (unsigned char *)isk_y_buffer, sizeof(isk_y_buffer));

    // A = B * (1/y)
    BIG_256_28_invmodp(isk_y_modInv, isk_y, CURVE_Order_FP256BN);
    ECP_FP256BN_copy(&A, &B);          // A=B
    ECP_FP256BN_mul(&A, isk_y_modInv); // A=A*(1/y)

    // C = (A + Q) * x
    LOG_I("C = (A + Q)^x");
    ECP_FP256BN_copy(&C, &A);
    ECP_FP256BN_add(&C, &Q);
    ECP_FP256BN_mul(&C, isk_x);

    // D = Q
    LOG_I("D = Q");
    ECP_FP256BN_copy(&D, &Q);

    LOG_I(
        "EcdaaJoin 9.3) The ECDAA Issuer sends A, CA,C to the authenticator. The authenticator still knows BB and DD");

    LOG_I(
        "EcdaaJoin 10) The authenticator store the credential on the SE (in real use case these are injected via a "
        "confidentiality protected mechanism)");
    // Copy A,C,D to buffer
    ecp_FP256BN_serialize((uint8_t *)A_buffer, &A);
    ecp_FP256BN_serialize((uint8_t *)C_buffer, &C);
    ecp_FP256BN_serialize((uint8_t *)D_buffer, &D);
    LOG_MAU8_I("A = B^(1/y)", (unsigned char *)A_buffer, sizeof(A_buffer));
    LOG_MAU8_I("C = (A + Q)^x", (unsigned char *)C_buffer, sizeof(C_buffer));
    LOG_MAU8_I("D = Q", (unsigned char *)D_buffer, sizeof(D_buffer));

    // Write key A
    status = sss_key_object_init(&pk_A, &pCtx->ks);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_allocate_handle(&pk_A,
        MAKE_TEST_ID(__LINE__),
        kSSS_KeyPart_Public,
        kSSS_CipherType_EC_BARRETO_NAEHRIG,
        keylen,
        kKeyObject_Mode_Persistent);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_store_set_key(&pCtx->ks, &pk_A, (uint8_t *)A_buffer, sizeof(A_buffer), KEY_BIT_LENGTH, NULL, 0);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    // Write key C
    status = sss_key_object_init(&pk_C, &pCtx->ks);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_allocate_handle(&pk_C,
        MAKE_TEST_ID(__LINE__),
        kSSS_KeyPart_Public,
        kSSS_CipherType_EC_BARRETO_NAEHRIG,
        keylen,
        kKeyObject_Mode_Persistent);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_store_set_key(&pCtx->ks, &pk_C, (uint8_t *)C_buffer, sizeof(C_buffer), KEY_BIT_LENGTH, NULL, 0);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    // Write key D
    status = sss_key_object_init(&pk_D, &pCtx->ks);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_allocate_handle(&pk_D,
        MAKE_TEST_ID(__LINE__),
        kSSS_KeyPart_Public,
        kSSS_CipherType_EC_BARRETO_NAEHRIG,
        keylen,
        kKeyObject_Mode_Persistent);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_store_set_key(&pCtx->ks, &pk_D, (uint8_t *)D_buffer, sizeof(D_buffer), KEY_BIT_LENGTH, NULL, 0);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    LOG_I(
        "EcdaaJoin operation completed!The SE and Backend established trust, from now on the SE can generate anonymous "
        "sigantures which the Backend can verify.");

    LOG_I("ECDAAsign PRE) Generate KRD");
    // keyRegistrationData(KRD) is 100Byte random
    status = sss_rng_context_init(&ctx_rng, &pCtx->session);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    status = sss_rng_get_random(&ctx_rng, (uint8_t *)keyRegistrationData, sizeof(keyRegistrationData));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    sss_rng_context_free(&ctx_rng);
    LOG_MAU8_I("KRD", (unsigned char *)keyRegistrationData, sizeof(keyRegistrationData));

    // Write key S
    status = sss_key_object_init(&pk_S, &pCtx->ks);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_allocate_handle(&pk_S,
        MAKE_TEST_ID(__LINE__),
        kSSS_KeyPart_Public,
        kSSS_CipherType_EC_BARRETO_NAEHRIG,
        keylen,
        kKeyObject_Mode_Transient);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    S_buffer[0] = 0x04;
    status = sss_key_store_set_key(&pCtx->ks, &pk_S, (uint8_t *)S_buffer, sizeof(S_buffer), KEY_BIT_LENGTH, NULL, 0);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    // Generate private key l
    status = sss_key_object_init(&sk_l, &pCtx->ks);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    status = sss_key_object_allocate_handle(&sk_l,
        MAKE_TEST_ID(__LINE__),
        kSSS_KeyPart_Private,
        kSSS_CipherType_EC_BARRETO_NAEHRIG,
        keylen,
        kKeyObject_Mode_Transient);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    /*
    status = sss_key_store_generate_key(&pCtx->ks, &sk_l, KEY_BIT_LENGTH, 0);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
*/
    sm_status = Se05x_API_WriteECKey(&se05x_session->s_ctx, // sss API?
        NULL,
        SE05x_MaxAttemps_NA,
        sk_l.keyId,
        kSE05x_ECCurve_TPM_ECC_BN_P256,
        NULL,
        0,
        NULL,
        0,
        kSE05x_INS_TRANSIENT,
        kSE05x_KeyPart_Private);
    ENSURE_OR_GO_CLEANUP(sm_status == SM_OK);

    LOG_I("ECDAAsign Note: From here on the complete operation must be repeatable without NV writes.");
    LOG_I("ECDAAsign a) regenerate random l");

    // l private key is 32Byte random mod p
    status = sss_rng_context_init(&ctx_rng, &pCtx->session);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    status = sss_rng_get_random(&ctx_rng, (uint8_t *)l_buffer, sizeof(l_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    sss_rng_context_free(&ctx_rng);

    BIG_256_28_fromBytes(l_BIG, (char *)l_buffer);
    BIG_256_28_mod(l_BIG, CURVE_Order_FP256BN);
    BIG_256_28_toBytes((char *)l_buffer, l_BIG);
    LOG_MAU8_I("l", (unsigned char *)l_buffer, sizeof(l_buffer));

    // Write l private key
    status = sss_key_store_set_key(&pCtx->ks, &sk_l, (uint8_t *)l_buffer, sizeof(l_buffer), KEY_BIT_LENGTH, NULL, 0);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    //Calculate R = ECPointMultiply(l, A)
    sm_status = Se05x_API_ECPointMultiply_InputObj(&se05x_session->s_ctx,
        sk_l.keyId,
        pk_A.keyId,
        0,
        (uint8_t *)R_buffer,
        &R_buffer_len,
        kSE05x_ECPMAlgo_SVDP_DH_PLAIN_XY);
    ENSURE_OR_GO_CLEANUP(sm_status == SM_OK);
    LOG_I("R = ECPointMultiply(l, A)");

    //Calculate S = ECPointMultiply(l, B)
    sm_status = Se05x_API_ECPointMultiply_InputObj(
        &se05x_session->s_ctx, sk_l.keyId, pk_host_b.keyId, pk_S.keyId, NULL, NULL, kSE05x_ECPMAlgo_SVDP_DH_PLAIN_XY);
    ENSURE_OR_GO_CLEANUP(sm_status == SM_OK);
    LOG_I("S = ECPointMultiply(l, B)");

    status = sss_key_store_get_key(&pCtx->ks, &pk_S, (uint8_t *)S_buffer, &s_key_byte_len, &s_key_bit_len);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    //Calculate T = ECPointMultiply(l, C)
    sm_status = Se05x_API_ECPointMultiply_InputObj(&se05x_session->s_ctx,
        sk_l.keyId,
        pk_C.keyId,
        0,
        (uint8_t *)T_buffer,
        &T_buffer_len,
        kSE05x_ECPMAlgo_SVDP_DH_PLAIN_XY);
    ENSURE_OR_GO_CLEANUP(sm_status == SM_OK);
    LOG_I("T = ECPointMultiply(l, C)");

    //Calculate W = ECPointMultiply(l, D)
    sm_status = Se05x_API_ECPointMultiply_InputObj(&se05x_session->s_ctx,
        sk_l.keyId,
        pk_D.keyId,
        0,
        (uint8_t *)W_buffer,
        &W_buffer_len,
        kSE05x_ECPMAlgo_SVDP_DH_PLAIN_XY);
    ENSURE_OR_GO_CLEANUP(sm_status == SM_OK);
    LOG_I("W = ECPointMultiply(l, D)");

    LOG_I("ECDAAsign b) Random (r) was already regenerated during the last sign operation.");
    /* We don't explicitly regenerate the value again, to not cause flash writes. */

    LOG_I("ECDAAsign c) Calculation of U on SE");
    //Calculate U = ECPointMultiply(r, S)
    sm_status = Se05x_API_ECPointMultiply_InputObj(&se05x_session->s_ctx,
        random_key.keyId,
        pk_S.keyId,
        0,
        (uint8_t *)U_buffer,
        &U_buffer_len,
        kSE05x_ECPMAlgo_SVDP_DH_PLAIN_XY);
    ENSURE_OR_GO_CLEANUP(sm_status == SM_OK);
    LOG_I("U = ECPointMultiply(r, S)");

    LOG_I("ECDAAsign d) Calculation of the Hash (c) on the Host");
    // c2 = H(U|S|W|AppID|H(KRD))
    // H(KRD) = SHA256(KRD) mod p
    status = H(hash_KRD_BIG, (uint8_t *)hash_KRD_buffer, (uint8_t *)keyRegistrationData, sizeof(keyRegistrationData));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("H(KRD) = H(KRD)", (unsigned char *)hash_KRD_buffer, sizeof(hash_KRD_buffer));

    LOG_I("AppID %s", appID);

    memcpy(&pre_c2_1_buffer[0], U_buffer, sizeof(U_buffer));
    memcpy(&pre_c2_1_buffer[sizeof(U_buffer)], S_buffer, sizeof(S_buffer));
    memcpy(&pre_c2_1_buffer[sizeof(U_buffer) + sizeof(S_buffer)], W_buffer, sizeof(W_buffer));
    memcpy(&pre_c2_1_buffer[sizeof(U_buffer) + sizeof(S_buffer) + sizeof(W_buffer)],
        appID,
        strlen(appID)); // Skip '\0' at the end of appID
    memcpy(&pre_c2_1_buffer[sizeof(U_buffer) + sizeof(S_buffer) + sizeof(W_buffer) + strlen(appID)],
        hash_KRD_buffer,
        sizeof(hash_KRD_buffer));
    status = H(c2_1_BIG, (uint8_t *)c2_1_buffer, (uint8_t *)pre_c2_1_buffer, sizeof(pre_c2_1_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("c2 = H(U|S|W|AppID|H(KRD))", (unsigned char *)c2_1_buffer, sizeof(c2_1_buffer));

    //c = H(n|c_2)

    // n is 32B random mod p
    status = sss_rng_context_init(&ctx_rng, &pCtx->session /* Session */);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    status = sss_rng_get_random(&ctx_rng, n_buffer, sizeof(n_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    sss_rng_context_free(&ctx_rng);

    BIG_256_28_fromBytes(n, (char *)n_buffer);
    BIG_256_28_mod(n, CURVE_Order_FP256BN);
    BIG_256_28_toBytes((char *)n_buffer, n);
    LOG_MAU8_I("n", (unsigned char *)n_buffer, sizeof(n_buffer));

    // n|c_2
    memcpy(&pre_c_buffer[0], n_buffer, sizeof(n_buffer));
    memcpy(&pre_c_buffer[sizeof(n_buffer)], c2_1_buffer, sizeof(c2_1_buffer));

    //c = H(n|c_2) = SHA256(n|c_2) mod p
    status = H(c_BIG, (uint8_t *)c_buffer, (uint8_t *)pre_c_buffer, sizeof(pre_c_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("c = H(n|c_2)", (unsigned char *)c_buffer, sizeof(c_buffer));

    LOG_I("ECDAAsign e) Calculation of the ECDAA signature on the SE");
    /* Asymmetric sign */
    status = sss_asymmetric_context_init(&ctx_asymm, &pCtx->session, &keyPair_sk, kAlgorithm_SSS_ECDAA, kMode_SSS_Sign);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);

    signatureLen = sizeof(signature);
    /* Do Signing */
    LOG_I("Do Signing");
    LOG_MAU8_I("digest", (uint8_t *)c_buffer, sizeof(c_buffer));
    status = sss_asymmetric_sign_digest(&ctx_asymm, (uint8_t *)c_buffer, sizeof(c_buffer), signature, &signatureLen);
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("signature", signature, signatureLen);
    LOG_I("Signing Successful !!!");
    sss_asymmetric_context_free(&ctx_asymm);
    // Raw signature is last 32Bytes of signature
    ENSURE_OR_GO_CLEANUP(signatureLen >= sizeof(s_buffer));
    memcpy(&s_buffer[0], &signature[signatureLen - sizeof(s_buffer)], sizeof(s_buffer));
    LOG_MAU8_I("Raw signature", (uint8_t *)s_buffer, sizeof(s_buffer));
    BIG_256_28_fromBytes(s_BIG, (char *)s_buffer);

    LOG_I("ECDAAsign f) Verify the ECDAA signature");
    ret = point_on_G1_not_infinity(R_buffer, sizeof(R_buffer), &R);
    ENSURE_OR_GO_CLEANUP(ret == 1); // Point R Is On G1 and Not Infinity
    LOG_I("Point R Is On G1 and Not Infinity");
    ret = point_on_G1_not_infinity(S_buffer, sizeof(S_buffer), &S);
    ENSURE_OR_GO_CLEANUP(ret == 1); // Point S Is On G1 and Not Infinity
    LOG_I("Point S Is On G1 and Not Infinity");
    ret = point_on_G1_not_infinity(T_buffer, sizeof(T_buffer), &T);
    ENSURE_OR_GO_CLEANUP(ret == 1); // Point T Is On G1 and Not Infinity
    LOG_I("Point T Is On G1 and Not Infinity");
    ret = point_on_G1_not_infinity(W_buffer, sizeof(W_buffer), &W);
    ENSURE_OR_GO_CLEANUP(ret == 1); // Point W Is On G1 and Not Infinity
    LOG_I("Point W Is On G1 and Not Infinity");

    ECP_FP256BN_copy(&S_s, &S);   // S_s = S
    ECP_FP256BN_mul(&S_s, s_BIG); // S_s = S * s
    ecp_FP256BN_serialize((uint8_t *)S_s_buffer, &S_s);
    LOG_MAU8_I("S^s", (uint8_t *)S_s_buffer, sizeof(S_s_buffer));

    ECP_FP256BN_copy(&W_c, &W);   // W_c = W
    ECP_FP256BN_mul(&W_c, c_BIG); // W_c = W * c
    ecp_FP256BN_serialize((uint8_t *)W_c_buffer, &W_c);
    LOG_MAU8_I("W^c", (uint8_t *)W_c_buffer, sizeof(W_c_buffer));

    ECP_FP256BN_copy(&U_verify, &S_s); // U_verify = S_s
    ECP_FP256BN_sub(&U_verify, &W_c);  // U_verify = S_s - W_c
    ecp_FP256BN_serialize((uint8_t *)U_verify_buffer, &U_verify);
    LOG_MAU8_I("U_verify=S^s-W^c", (uint8_t *)U_verify_buffer, sizeof(U_verify_buffer));

    // U_verify|S|W|AppID|H(KRD)
    memcpy(&pre_c_2_2_verify_buffer[0], U_verify_buffer, sizeof(U_verify_buffer));
    memcpy(&pre_c_2_2_verify_buffer[sizeof(U_verify_buffer)], S_buffer, sizeof(S_buffer));
    memcpy(&pre_c_2_2_verify_buffer[sizeof(U_verify_buffer) + sizeof(S_buffer)], W_buffer, sizeof(W_buffer));
    memcpy(&pre_c_2_2_verify_buffer[sizeof(U_verify_buffer) + sizeof(P1_256_buffer) + sizeof(W_buffer)],
        appID,
        strlen(appID));
    memcpy(&pre_c_2_2_verify_buffer[sizeof(U_verify_buffer) + sizeof(P1_256_buffer) + sizeof(W_buffer) + strlen(appID)],
        hash_KRD_buffer,
        sizeof(hash_KRD_buffer));

    //c_2_verify = H(U_verify|S|W|AppID|H(KRD)) = SHA256(U_verify|S|W|AppID|H(KRD)) mod p
    status = H(c_2_2_verify_BIG,
        (uint8_t *)c_2_2_verify_buffer,
        (uint8_t *)pre_c_2_2_verify_buffer,
        sizeof(pre_c_2_2_verify_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I(
        "c_2_verify = H(U_verify|S|W|AppID|H(KRD))", (unsigned char *)c_2_2_verify_buffer, sizeof(c_2_2_verify_buffer));

    // n|c_2_verify
    memcpy(&pre_c_verify_buffer[0], n_buffer, sizeof(n_buffer));
    memcpy(&pre_c_verify_buffer[sizeof(n_buffer)], c_2_2_verify_buffer, sizeof(c_2_2_verify_buffer));

    //c_verify = H(n|c_2_verify) = SHA256(n|c_2_verify) mod p
    status = H(c_verify_BIG, (uint8_t *)c_verify_buffer, (uint8_t *)pre_c_verify_buffer, sizeof(pre_c_verify_buffer));
    ENSURE_OR_GO_CLEANUP(status == kStatus_SSS_Success);
    LOG_MAU8_I("c_verify = H(n|c_2_verify)", (unsigned char *)c_verify_buffer, sizeof(c_verify_buffer));
    LOG_MAU8_I("c", (unsigned char *)c_buffer, sizeof(c_buffer));

    ENSURE_OR_GO_CLEANUP(memcmp(c_verify_buffer, c_buffer, sizeof(c_buffer)) == 0);

cleanup:
    if (kStatus_SSS_Success == status) {
        LOG_I("se05x_FIDO_ECDAA Example Success !!!...");
    }
    else {
        LOG_E("se05x_FIDO_ECDAA Example Failed !!!...");
    }
    if (ctx_asymm.session != NULL) {
        sss_asymmetric_context_free(&ctx_asymm);
    }
    return status;
}
